"""
PCA and MNF methods for dimensionality reduction.
"""

import numpy as np

import spectral
from hylite import HyData

def PCA(hydata, bands=20, band_range=None, step=5, mask : np.ndarray = None):
    """
    Apply a PCA dimensionality reduction to the hyperspectral dataset using singular vector decomposition (SVD).

    Args:
        data: the dataset (HyData object) to apply PCA to.
        output_bands: number of bands to return (i.e. how many dimensions to retain). Default is 20.
        bands: the spectral range to perform the PCA over. If (int,int) is passed then the values are treated as
                    min/max band IDs, if (float,float) is passed then values are treated as wavelenghts (in nm). If None is
                    passed (default) then the PCA is computed using all bands. Note that wavelengths can only be passed
                    if image is a hyImage object.
        step: subsample the dataset during SVD for performance reason. step = 1 will include all pixels in the calculation,
              step = n includes every nth pixel only. Default is 5 (as most images contain more than enough pixels to
              accurately estimate variance etc.).
        mask: A mask containing pixels to ignore during the fitting of this PCA. Default is None (consider all pixels).
    Returns:
        A tuple containing:

        - bands = A HyData instance containing the PCA components, ordered from highest to lowest variance.
        - factors = the factors (vector) each band is multiplied with to give the corresponding PCA band.
        - means = the means for each band that are subtracted before applying the PCA transform.

        Additional info (including loadings and the per-band means) are stored in the header file of the returned HyData
        instance.
    """

    # get numpy array
    wav = None
    decomp = False
    if isinstance(hydata, HyData):
        wav = hydata.get_wavelengths()

        if hydata.is_int():
            hydata.decompress()  # MNF doesn't work very well with ints....
            decomp = True  # so we can compress again afterwards

        data = hydata.data.copy()
    else:
        data = hydata.copy()

    # get band range
    if band_range is None:  # default to all bands
        minb = 0
        maxb = data.shape[-1]
    else:
        if isinstance(band_range[0], int) and isinstance(band_range[1], int):
            minb, maxb = band_range
        else:
            assert isinstance(hydata, HyData), "Error - no wavelength information found."
            minb = hydata.get_band_index(band_range[0])
            maxb = hydata.get_band_index(band_range[1])

    # prepare feature vectors
    if mask is not None:
        X = data[mask, : ]
    else:
        X = data[..., :].reshape(-1, data.shape[-1])

    # print(minb,maxb)
    X = X[::step, minb:maxb]  # subsample
    X = X[np.isfinite(np.sum(X, axis=1)), :]  # drop vectors containing nans
    X = X[np.sum(X, axis=1) > 0, :]  # drop vectors containing all zeros

    # calculate mean and center
    mean = np.mean(X, axis=0)
    X = X - mean[None, :]

    # calculate covariance
    cov = np.dot(X.T, X) / (X.shape[0] - 1)

    # and eigens (sorted from biggest to smallest)
    eigval, eigvec = np.linalg.eig(cov)
    idx = np.argsort(np.abs(eigval))[::-1]
    eigvec = eigvec[:, idx]
    eigval = np.abs(eigval[idx])

    # project data
    data = data[..., minb:maxb] - mean
    #out = np.zeros_like(data)
    #for b in range(min(bands, data.shape[-1])):
    #    out[..., b] = np.dot(data, eigvec[:, b])
    out = np.dot(data, eigvec)

    # compute variance percentage of each eigenvalue
    eigval /= np.sum(eigval)  # sum to 1

    # compress?
    if decomp:
        hydata.compress()

    # prepare output
    bands = min(bands, out.shape[-1])
    if isinstance(hydata, HyData):
        outobj = hydata.copy(data=False)
        outobj.data = out[..., 0:bands]
    else:
        outobj = HyData( out[..., 0:bands] )
    outobj.set_wavelengths(np.cumsum(eigval[0:bands]))  # wavelengths are % of explained variance
    outobj.header['mean'] = mean
    for n in range(bands):
        outobj.header['L_%d'%n] = eigvec[:, n]
    return outobj, eigvec[:, :bands ].T, mean


def MNF(hydata, bands=20, band_range=None, noise='diff', noise_thresh=50, denoise=False, mask : np.ndarray =None):
    """
    Apply a minimum noise filter to a hyperspectral image.

    Args:
        hydata: A HyData instance containing the source dataset (e.g. image or point cloud).
        bands: the number of bands to keep after MNF for dimensionality reduction or denoising. Default is 20.
        band_range: the spectral range to perform the MNF over. If (int,int) is passed then the values are treated as
                    min/max band IDs, if (float,float) is passed then values are treated as wavelenghts (in nm). If None is
                    passed (default) then the MNF is computed using all bands. Note that wavelengths can only be passed
                    if image is a hyImage object.
        noise: The noise model to use. If None (default) then the 'band noise' parameter will be retrieved from the headerinfo,
                and if this does not exist then, this it is crudely estimated by comparing adjacent pixels / points.
                This estimation can be forced by setting noise to 'diff'.
        noise_thresh: The threshold percentile used when estimating noise with 'diff'.
        denoise: True if a MNF denoised image should be returned (rather than the MNF bands). Default is False.
        mask: A boolean mask containing pixels to ignore during the fitting of this PCA. Default is None (consider all pixels).
    Returns:
        A tuple containing:

        - mnf = a HyData instance containing the MNF bands or denoised image.
        - factors = A 2D numpy array containing the factors applied to the input datset. Useful
                     for plotting/interpreting the regions each MNF band is sensitive too.
        - means = the means for each band that are subtracted before applying the MNF transform.
    """

    # prepare data for MNF
    wav = hydata.get_wavelengths()
    decomp = False
    if hydata.is_int():
        hydata.decompress() #MNF doesn't work very well with ints....
        decomp=True #so we can compress again afterwards
    data = hydata.data

    # get range of bands to include in calculation
    if band_range is None:  # default to all bands
        minb = 0
        maxb = data.shape[-1]
    else:
        minb = hydata.get_band_index( band_range[0] )
        maxb = hydata.get_band_index( band_range[1] )

    assert minb < maxb, "Error - invalid range... band_range[0] > band_range[1]??"
    assert minb < data.shape[-1], "Error - band_range[0] out of range."
    if maxb == -1 or maxb > data.shape[-1]: maxb = data.shape[-1]

    # remove invalid bands
    valid_bands = []
    for b in range(minb,maxb):
        if np.isfinite(data[..., b]).any() \
                and not (np.nanmax(data[..., b]) == 0).all():
            valid_bands.append(b)

    #remove invalid bands
    data = np.array(data[..., valid_bands])

    # warn if bands have negative values...
    if np.nanmin(data) < 0.0:
        print("Warning - image contains negative pixels. This can cause unstable behaviour...")

    # calculate signal stats (as in spectral.calc_stats(...) but allowing for nans)
    if mask is not None:
        X = data[mask, : ].T
    else:
        X = data.reshape(-1, data.shape[-1]).T  # reshape to 1D list of pixels for each band

    X = X[:, np.isfinite(np.sum(X, axis=0))]  # drop columns containing nans
    X = X[:, np.sum(X, axis=0) > 0 ] #drop columns containing all zeros
    dmean = np.mean(X, axis = 1)
    cov = np.cov(X)
    n = X.shape[1]
    signal = spectral.GaussianStats(dmean, cov, n)

    if (noise is None) and 'band noise' in hydata.header:  # get noise from header?
        assert False, "Error - this is not yet implemented! [ hopefully soon! ]"
        noise = np.array( hydata.header.get_list('band noise') )
    if (noise is not None) and (noise != 'diff'): # we have noise info - use it
        assert False, "Error - this is not yet implemented! [ hopefully soon! ]"
        assert noise.shape[0] == hydata.band_count(), "Error - noise provided for %d bands but data has %d" % (noise.shape[0], hydata.band_count())
        mean = noise[valid_bands] # easy
        cov = np.full( (mean.shape[0], mean.shape[0]), 0. ) # assume noise is de-corellated
        noise = spectral.GaussianStats(mean, cov, n)
    else: # we need to guesstimate the noise model from the data
        if len(data.shape) == 3:  # estimate noise by subtracting adjacent pixels
            if data.shape[0] == 1:  # special case - 1-D image
                deltas = data[:, :-1, :] - data[:, 1:, :]
            elif data.shape[1] == 1:
                deltas = data[:-1, :, :] - data[1:, :, :]
            else:
                deltas = data[:-1, :-1, :] - data[1:, 1:, :]
        elif len(data.shape) == 2:  # point cloud data. N.B. this assumes that adjacent points are close in space!
            deltas = data[:-1, :] - data[1:, :]

        X = deltas.reshape(-1, deltas.shape[-1]).T
        X = X[:, np.isfinite(np.sum(X, axis=0))]  # drop columns containing nans
        X = X[:, np.sum(X, axis=0) > 0]  # drop columns containing all zeros
        X = X[:, np.sum(X, axis=0) < np.nanpercentile( np.sum(X,axis=0), noise_thresh ) ] # drop high noise data (these relate to edges)
        assert np.isfinite(X).all(), "Error - the nans snuck into our data!"
        mean = np.mean(X, axis=1)
        cov = np.cov(X)
        #n = X.shape[1]

        noise = spectral.GaussianStats(mean, cov, n)
    mnfr = spectral.mnf(signal, noise)

    # reduce bands
    reduced = mnfr.reduce(data, num=bands)

    #apply sign correction so there are less positive pixels than negative ones (sign is aribrary, helps maintain
    #consistency for plotting etc. by having low-valued background with some high-value regions (<50%)
    sign = np.nanmedian(reduced / np.abs(reduced)) #n.b. this will always be 1.0 or -1.0
    assert np.isfinite(sign), "Weird error - no non-nan values in MNF result?"
    reduced *= sign

    # calculate factors for return
    factors = sign * mnfr.get_reduction_transform(num=bands)._A

    # denoise and export
    if denoise:
        out = hydata.copy(data=False)
        out.data = mnfr.denoise(data, num=bands)
        out.set_wavelengths(hydata.get_wavelengths()[valid_bands])
    else:
        out = hydata.copy(data=False)
        out.header.drop_all_bands()  # drop band specific attributes
        out.data = reduced.reshape(hydata.data.shape[:-1] + (bands,))
        out.push_to_header()

    # compress input dataset (so we don't change it)
    if decomp:
        hydata.compress()

    # put factors in the header
    for i in range(factors.shape[0]):
        out.header['mnf weights %d' % i] = factors[i, :]

    return out, factors, dmean


def from_loadings(data, L, m=None):
    """
    Transform a dataset using a precomputed loading vector.  This allows PCA
    transforms to be computed on one dataset and then applied to another.

    Args:
       data: A dataset (HyData instance or numpy array) with b bands in the last axis.
       L: the loadings vector of shape (k,b), such that data is projected into a
                 k-dimensional space.
       m: the mean of each dimension. Default is None ( do not apply mean offset ).
    Returns:
       a hydata instance or numpy array containing the transformed data.
    """
    # get relevant data
    if isinstance(data, HyData):
        X = data.X()
        outshape = data.data.shape[:-1] + (L.shape[0],)
    else:
        X = data.reshape((-1, data.shape[-1]))
        outshape = data.shape[:-1] + (L.shape[0],)

    # project data
    if m is not None:
        X = X - m
    out = np.dot(X, L.T)

    # add mean
    #if m is not None:
    #    out += m

    # reshape
    out = out.reshape(outshape)

    # return HyData or numpy array
    if isinstance(data, HyData):
        O = data.copy(data=False)
        O.data = out
        return O
    else:
        return out



"""
Implement decision tree function for classification. This is done in a generic form using decision_tree( ... ), but also
for specific minerals based on minimum wavelength data in subsequent functions.
"""

import numpy as np

def decision_tree( layers, labels ):
    """
    Implement a simple binary tree for assigning classes based on boolean input layers.

    Args:
        layers (list): a list of boolean arrays indicating the value of each pixel/point at each successive layer in the tree.
        labels (dict): a dictionary keyed by leaves of the decision tree (tuples containing a boolean value for each level in
                the tree) with values that give class labels. Leaves that are not in the tree will be assigned
                a label of 0 (unknown). For example: { (True,False,True) : 1, (False,False,True) : 2 } will label the
                leaves of the tree True -> False -> True as 1 and False -> False -> True as 2, and all others as 0. Use
                None to skip branches for certain labels (e.g. True -> None -> False will ignore the second layer of
                the decision tree for this label).

    Returns:
         an output labels array of the sampe shape as each layer array but containing 0 ... n class labels.
    """

    out = np.zeros( layers[0].shape, dtype=int )
    names = ["Unknown"]
    for k in labels.items():
        key, value = k
        if not value in names:
            names.append(value)
        msk = np.full( layers[0].shape, True )
        for n,l in enumerate(layers):
            if n < len(key):
                if key[n] is not None:
                    msk = msk & (l == key[n])
        out[ msk ]=names.index(value)
    return out, names






"""
A simple data structure for storing camera pose information.
"""

import numpy as np
from scipy import spatial

class Camera( object ):
    """
    A utility class for storing camera properties (position, dimensions, fov etc.).
    """

    def __init__(self, pos, ori, proj, fov, dims, step = None):
        """
        Args:
            pos (ndarray): the camera position vector.
            ori (ndarray): the camera orientation vector (pitch, roll, yaw) in degrees.
            proj (str): the project type: 'persp' or 'pano'.
            fov (float): the camera field of view (degrees).
            dims (tuple): the pixel dimensions of the frame.
            step (float): the angular step of pixels in the x-direction for panoramic projections. Default is None.
        """

        assert isinstance(pos, np.ndarray) and len(pos)==3, \
                                            "Error - camera_position must be a numpy array with length 3. "
        self.pos = pos.copy()

        assert isinstance(ori, np.ndarray) and len(ori) == 3, \
                                            "Error - camera_ori must be a numpy array with length 3. "
        self.ori = ori.copy()

        assert 'persp' in proj.lower() or 'pano' in proj.lower(), \
                           "Error - camera type should be pano or persp."
        self.proj = proj

        assert len(dims) >= 2, "Error, dims must be a tuple or list of length 2."
        self.dims = dims

        assert np.isfinite(fov), "Error, fov must be numeric."
        self.fov = fov
        self.step = step

    def clone(self):
        """
        Create a deep copy of a camera instance.
        """
        return Camera( self.pos.copy(), self.ori.copy(), self.proj, self.fov, self.dims, self.step )

    def set_transform(self, camera_pos, camera_ori):
        """
        Sets the position and rotation of this camera.

        Args:
            camera_pos (ndarray): the new position.
            camera_ori (ndarray): the new orientation.
        """
        assert isinstance(camera_pos, np.ndarray) and len(camera_pos) == 3, \
                                        "Error - camera_position must be a numpy array with length 3. "
        self.pos = camera_pos.copy()

        assert isinstance(camera_ori, np.ndarray) and len(camera_ori) == 3, \
                                        "Error - camera_ori must be a numpy array with length 3. "
        self.ori = camera_ori.copy()

    def xdim(self):
        """
        Return camera x-dimension (pixels).
        """
        return self.dims[0]

    def ydim(self):
        """
        Return camera y-dimension (pixels).
        """
        return self.dims[1]

    def is_perspective(self):
        """
        Returns true if this camera uses a perspective project.
        """
        return 'persp' in self.proj.lower()

    def is_panoramic(self):
        """
        Returns true if this camera uses a panoramic project.
        """
        return 'pano' in self.proj.lower()

    def get_rotation_matrix(self):
        """
        Return the rotation matrix of this camera (based on its Euler angles).
        """
        return spatial.transform.Rotation.from_euler('XYZ', -self.ori, degrees=True).as_matrix()



"""
Project points between 2D image coordinates and 3D world coordinates. Also includes related problems such as
camera localisation.
"""

from .basic import *
from .camera import Camera
from .pmap import PMap, push_to_cloud, push_to_image, blend_scenes, push_geomattr, get_blend_weights
from .pushbroom import Pushbroom, project_pushbroom, optimize_boresight


"""
Store and manipulate hyperspectral image data.
"""

import os
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import path
from roipoly import MultiRoi
import imageio
import cv2
import scipy as sp
from scipy import ndimage
import hylite
from hylite.hydata import HyData
from hylite.hylibrary import HyLibrary



class HyImage( HyData ):
    """
    A class for hyperspectral image data. These can be individual scenes or hyperspectral orthoimages.
    """

    def __init__(self, data, **kwds):
        """
        Args:
            data (ndarray): a numpy array such that data[x][y][band] gives each pixel value.
            **kwds:
                wav = A numpy array containing band wavelengths for this image.
                affine = an affine transform of the format returned by GDAL.GetGeoTransform().
                projection = string defining the project. Default is None.
                sensor = sensor name. Default is "unknown".
                header = path to associated header file. Default is None.
        """

        #call constructor for HyData
        super().__init__(data, **kwds)

        # special case - if dataset only has oneband, slice it so it still has
        # the format data[x,y,b].
        if not self.data is None:
            if len(self.data.shape) == 1:
                self.data = self.data[None, None, :] # single pixel image
            if len(self.data.shape) == 2:
                self.data = self.data[:, :, None] # single band iamge

        #load any additional project information (specific to images)
        self.set_projection(kwds.get("projection",None))
        self.affine = kwds.get("affine",[0,1,0,0,0,1])

        # wavelengths
        if 'wav' in kwds:
            self.set_wavelengths(kwds['wav'])

        #special header formatting
        self.header['file type'] = 'ENVI Standard'

    def copy(self,data=True):
        """
        Make a deep copy of this image instance.

        Args:
            data (bool): True if a copy of the data should be made, otherwise only copy header.

        Returns:
            a new HyImage instance.
        """
        if not data:
            return HyImage(None, header=self.header.copy(), projection=self.projection, affine=self.affine)
        else:
            return HyImage( self.data.copy(), header=self.header.copy(), projection=self.projection, affine=self.affine)

    def T(self):
        """
        Return a transposed view of the data matrix (corresponding with the [y,x] indexing used by matplotlib, opencv etc.
        """
        return np.transpose(self.data, (1,0,2))

    def xdim(self):
        """
        Return number of pixels in x (first dimension of data array)
        """
        return self.data.shape[0]

    def ydim(self):
        """
        Return number of pixels in y (second dimension of data array)
        """
        return self.data.shape[1]

    def aspx(self):
        """
        Return the aspect ratio of this image (width/height).
        """
        return self.ydim() / self.xdim()

    def get_extent(self):
        """
        Returns the width and height of this image in world coordinates.

        Returns:
            tuple with (width, height).
        """
        return self.xdim * self.pixel_size[0], self.ydim * self.pixel_size[1]

    def set_projection(self,proj):
        """
        Set this project to an existing osgeo.osr.SpatialReference or GDAL georeference string.

        Args:
            proj (str, osgeo.osr.SpatialReference): the project to use as osgeo.osr.SpatialReference or GDAL georeference string.
        """
        if proj is None:
            self.projection = None
        else:
            try:
                from osgeo.osr import SpatialReference
            except:
                assert False, "Error - GDAL must be installed to work with spatial projections in hylite."
            if isinstance(proj, SpatialReference):
                self.projection = proj
            elif isinstance(proj, str):
                self.projection = SpatialReference(proj)
            else:
                print("Invalid project %s" % proj)
                raise

    def set_projection_EPSG(self,EPSG):
        """
        Sets this image project using an EPSG code.

        Args:
            EPSG (str): string EPSG code that can be passed to SpatialReference.SetFromUserInput(...).
        """

        try:
            from osgeo.osr import SpatialReference
        except:
            assert False, "Error - GDAL must be installed to work with spatial projections in hylite."

        self.projection = SpatialReference()
        self.projection.SetFromUserInput(EPSG)

    def get_projection_EPSG(self):
        """
        Gets a string describing this projections EPSG code (if it is an EPSG project).

        Returns:
            an EPSG code string of the format "EPSG:XXXX".
        """
        if self.projection is None:
            return None
        else:
            return "%s:%s" % (self.projection.GetAttrValue("AUTHORITY",0),self.projection.GetAttrValue("AUTHORITY",1))

    def pix_to_world(self, px, py, proj=None):
        """
        Take pixel coordinates and return world coordinates

        Args:
            px (int): the pixel x-coord.
            py (int): the pixel y-coord.
            proj (str, osr.SpatialReference): the coordinate system to use. Default (None) uses the same system as this image. Otherwise
                   an osr.SpatialReference can be passed (HyImage.project), or an EPSG string (e.g. get_projection_EPSG(...)).
        Returns:
            the world coordinates in the coordinate system defined by get_projection_EPSG(...).
        """

        try:
            from osgeo import osr
            import osgeo.gdal as gdal
            from osgeo import ogr
        except:
            assert False, "Error - GDAL must be installed to work with spatial projections in hylite."

        # parse project
        if proj is None:
            proj = self.projection
        elif isinstance(proj, str) or isinstance(proj, int):
            epsg = proj
            if isinstance(epsg, str):
                try:
                    epsg = int(str.split(':')[1])
                except:
                    assert False, "Error - %s is an invalid EPSG code." % proj
            proj = osr.SpatialReference()
            proj.ImportFromEPSG(epsg)

        # check we have all the required info
        assert isinstance(proj, osr.SpatialReference), "Error - invalid spatial reference %s" % proj
        assert (not self.affine is None) and (
            not self.projection is None), "Error - project information is undefined."

        #project to world coordinates in this images project/world coords
        x,y = gdal.ApplyGeoTransform(self.affine, px, py)

        #project to target coords (if different)
        if not proj.IsSameGeogCS(self.projection):
            P = ogr.Geometry(ogr.wkbPoint)
            if proj.EPSGTreatsAsNorthingEasting():
                P.AddPoint(x, y)
            else:
                P.AddPoint(y, x)
            P.AssignSpatialReference(self.projection)  # tell the point what coordinates it's in
            P.TransformTo(proj)  # reproject it to the out spatial reference
            x, y = P.GetX(), P.GetY()

            #do we need to transpose?
            if proj.EPSGTreatsAsLatLong():
                x,y=y,x #we want lon,lat not lat,lon
        return x, y

    def world_to_pix(self, x, y, proj = None):
        """
        Take world coordinates and return pixel coordinates

        Args:
            x (float): the world x-coord.
            y (float): the world y-coord.
            proj (str, osr.SpatialReference): the coordinate system of the input coordinates. Default (None) uses the same system as this image. Otherwise
                   an osr.SpatialReference can be passed (HyImage.project), or an EPSG string (e.g. get_projection_EPSG(...)).

        Returns:
            the pixel coordinates based on the affine transform stored in self.affine.
        """

        try:
            from osgeo import osr
            import osgeo.gdal as gdal
            from osgeo import ogr
        except:
            assert False, "Error - GDAL must be installed to work with spatial projections in hylite."

        # parse project
        if proj is None:
            proj = self.projection
        elif isinstance(proj, str) or isinstance(proj, int):
            epsg = proj
            if isinstance(epsg, str):
                try:
                    epsg = int(str.split(':')[1])
                except:
                    assert False, "Error - %s is an invalid EPSG code." % proj
            proj = osr.SpatialReference()
            proj.ImportFromEPSG(epsg)


        # check we have all the required info
        assert isinstance(proj, osr.SpatialReference), "Error - invalid spatial reference %s" % proj
        assert (not self.affine is None) and (not self.projection is None), "Error - project information is undefined."

        # project to this images CS (if different)
        if not proj.IsSameGeogCS(self.projection):
            P = ogr.Geometry(ogr.wkbPoint)
            if proj.EPSGTreatsAsNorthingEasting():
                P.AddPoint(x, y)
            else:
                P.AddPoint(y, x)
            P.AssignSpatialReference(proj)  # tell the point what coordinates it's in
            P.AddPoint(x, y)
            P.TransformTo(self.projection)  # reproject it to the out spatial reference
            x, y = P.GetX(), P.GetY()
            if self.projection.EPSGTreatsAsLatLong(): # do we need to transpose?
                x, y = y, x  # we want lon,lat not lat,lon

        inv = gdal.InvGeoTransform(self.affine)
        assert not inv is None, "Error - could not invert affine transform?"

        #apply
        return gdal.ApplyGeoTransform(inv, x, y)

    def flip(self, axis='x'):
        """
        Flip the image on the x or y axis.

        Args:
            axis (str): 'x' or 'y' or both 'xy'.
        """

        if 'x' in axis.lower():
            self.data = np.flip(self.data,axis=0)
        if 'y' in axis.lower():
            self.data = np.flip(self.data,axis=1)

    def rot90(self):
        """
        Rotate this image by 90 degrees by transposing the underlying data array. Combine with flip('x') or flip('y')
        to achieve positive/negative rotations.
        """
        self.data = np.transpose( self.data, (1,0,2) )
        self.push_to_header()

    #####################################
    ##IMAGE FILTERING
    #####################################
    def fill_holes(self):
        """
        Replaces nan pixel with an average of their neighbours, thus removing 1-pixel large holes from an image. Note that
        for performance reasons this assumes that holes line up across bands. Note that this is not vectorized so very slow...
        """

        # perform greyscale dilation
        dilate = self.data.copy()
        mask = np.logical_not(np.isfinite(dilate))
        dilate[mask] = 0
        for b in range(self.band_count()):
            dilate[:, :, b] = sp.ndimage.grey_dilation(dilate[:, :, b], size=(3, 3))

        # map back to holes in dataset
        self.data[mask] = dilate[mask]
        #self.data[self.data == 0] = np.nan  # replace remaining 0's with nans

    def blur(self, n=3):
        """
        Applies a gaussian kernel of size n to the image using OpenCV.

        Args:
            n (int): the dimensions of the gaussian kernel to convolve. Default is 3. Increase for more blurry results.
        """

        nanmask = np.isnan(self.data)
        assert isinstance(n, int) and n >= 3, "Error - invalid kernel. N must be an integer > 3. "
        kernel = np.ones((n, n), np.float32) / (n ** 2)
        self.data = cv2.filter2D(self.data, -1, kernel)
        self.data[nanmask] = np.nan  # remove mask

    def erode(self, size=3, iterations=1):
        """
        Apply an erode filter to this image to expand background (nan) pixels. Refer to open-cv's erode
        function for more details.

        Args:
            size (int): the size of the erode filter. Default is a 3x3 kernel.
            iterations (int): the number of erode iterations. Default is 1.
        """

        # erode
        kernel = np.ones((size, size), np.uint8)
        if self.is_float():
            mask = np.isfinite(self.data).any(axis=-1)
            mask = cv2.erode(mask.astype(np.uint8), kernel, iterations=iterations)
            self.data[mask == 0, :] = np.nan
        else:
            mask = (self.data != 0).any( axis=-1 )
            mask = cv2.erode(mask.astype(np.uint8), kernel, iterations=iterations)
            self.data[mask == 0, :] = 0

    def resize(self, newdims, interpolation=cv2.INTER_LINEAR):
        """
        Resize this image with opencv.

        Args:
            newdims (tuple): the new image dimensions.
            interpolation (int): opencv interpolation method. Default is cv2.INTER_LINEAR.
        """

        self.data = cv2.resize(self.data, (newdims[1],newdims[0]), interpolation=interpolation)

    def despeckle(self, size=5):
        """
        Despeckle each band of this image (independently) using a median filter.

        Args:
            size (int): the size of the median filter kernel. Default is 5. Must be an odd number.
        """

        assert (size % 2) == 1, "Error - size must be an odd integer"

        if self.is_float():
            self.data = cv2.medianBlur( self.data.astype(np.float32), size )
        else:
            self.data = cv2.medianBlur( self.data, size )

    #####################################
    ##FEATURES AND FEATURE MATCHING
    ######################################
    def get_keypoints(self, band, eq=False, mask=True, method='sift', cfac=0.0,bfac=0.0, **kwds):
        """
        Get feature descriptors from the specified band.

        Args:
            band (int,float,str,tuple): the band index (int) or wavelength (float) to extract features from. Alternatively, a tuple can be passed
                    containing a range of bands (min : max) to average before feature matching.
            eq (bool): True if the image should be histogram equalized first. Default is False.
            mask (bool): True if 0 value pixels should be masked. Default is True.
            method (str): the feature detector to use. Options are 'SIFT' and 'ORB' (faster but less accurate). Default is 'SIFT'.
            cfac (float): contrast adjustment to apply to hyperspectral bands before matching. Default is 0.0.
            bfac (float): brightness adjustment to apply to hyperspectral bands before matching. Default is 0.0.
            **kwds: keyword arguments are passed to the opencv feature detector. For SIFT these are:

                - contrastThreshold: default is 0.01.
                - edgeThreshold: default is 10.
                - sigma: default is 1.0

                For ORB these are:

                - nfeatures = the number of features to detect. Default is 5000.

            Returns:
                Tuple containing

                    - k (ndarray): the keypoints detected
                    - d (ndarray): corresponding feature descriptors
         """

        # get image
        if isinstance(band, int) or isinstance(band, float): #single band
            image = self.data[:, :, self.get_band_index(band)]
        elif isinstance(band,tuple): #range of bands (averaged)
            idx0 = self.get_band_index(band[0])
            idx1 = self.get_band_index(band[1])

            #deal with out of range errors
            if idx0 is None:
                idx0 = 0
            if idx1 is None:
                idx1 = self.band_count()

            #average bands
            image = np.nanmean(self.data[:,:,idx0:idx1],axis=2)
        else:
            assert False, "Error, unrecognised band %s" % band

        #normalise image to range 0 - 1
        image -= np.nanmin(image)
        image = image / np.nanmax(image)

        #apply brightness/contrast adjustment
        image = (1.0+cfac)*image + bfac
        image[image > 1.0] = 1.0
        image[image < 0.0] = 0.0

        #convert image to uint8 for opencv
        image = np.uint8(255 * image)
        if eq:
            image = cv2.equalizeHist(image)

        if mask:
            mask = np.zeros(image.shape, dtype=np.uint8)
            mask[image != 0] = 255  # include only non-zero pixels
        else:
            mask = None

        if 'sift' in method.lower():  # SIFT

            # setup default keywords
            kwds["contrastThreshold"] = kwds.get("contrastThreshold", 0.01)
            kwds["edgeThreshold"] = kwds.get("edgeThreshold", 10)
            kwds["sigma"] = kwds.get("sigma", 1.0)

            # make feature detector
            #alg = cv2.xfeatures2d.SIFT_create(**kwds)
            alg = cv2.SIFT_create()
        elif 'orb' in method.lower():  # orb
            kwds['nfeatures'] = kwds.get('nfeatures', 5000)
            alg = cv2.ORB_create(scoreType=cv2.ORB_FAST_SCORE, **kwds)
        else:
            assert False, "Error - %s is not a recognised feature detector." % method

        # detect keypoints
        kp = alg.detect(image, mask)

        # extract and return feature vectors
        return alg.compute(image, kp)

    @classmethod
    def match_keypoints(cls, kp1, kp2, d1, d2, method='SIFT', dist=0.7, tree = 5, check = 100, min_count=5):
        """
        Compares keypoint feature vectors from two images and returns matching pairs.

        Args:
            kp1 (ndarray): keypoints from the first image
            kp2 (ndarray): keypoints from the second image
            d1 (ndarray): descriptors for the keypoints from the first image
            d2 (ndarray): descriptors for the keypoints from the second image
            method (str): the method used to calculate the feature descriptors. Should be 'sift' or 'orb'. Default is 'sift'.
            dist (float): minimum match distance (0 to 1), default is 0.7
            tree (int): not sure what this does? Default is 5. See open-cv docs.
            check (int): ditto. Default is 100.
            min_count (int): the minimum number of matches to consider a valid matching operation. If fewer matches are found,
                       then the function returns None, None. Default is 5.
        """

        if 'sift' in method.lower():
            algorithm = cv2.NORM_INF
        elif 'orb' in method.lower():
            algorithm = cv2.NORM_HAMMING
        else:
            assert False, "Error - unknown matching algorithm %s" % method

        #calculate flann matches
        index_params = dict(algorithm=algorithm, trees=tree)
        search_params = dict(checks=check)
        flann = cv2.FlannBasedMatcher(index_params, search_params)
        matches = flann.knnMatch(d1, d2, k=2)

        # store all the good matches as per Lowe's ratio test.
        good = []
        for m, n in matches:
            if m.distance < dist * n.distance:
                good.append(m)

        if len(good) < min_count:
            return None, None
        else:
            src_pts = np.float32([kp1[m.queryIdx].pt for m in good]).reshape(-1, 1, 2)
            dst_pts = np.float32([kp2[m.trainIdx].pt for m in good]).reshape(-1, 1, 2)
            return src_pts, dst_pts

    ############################
    ## Visualisation methods
    ############################
    def quick_plot(self, band=0, ax=None, bfac=0.0, cfac=0.0, samples=False, tscale=False, rot=False, flipX=False, flipY=False,
                   **kwds):
        """
        Plot a band using matplotlib.imshow(...).

        Args:
            band (str,int,float,tuple): the band name (string), index (integer) or wavelength (float) to plot. Default is 0. If a tuple is passed then
                  each band in the tuple (string or index) will be mapped to rgb. Bands with negative wavelengths or indices will be inverted before plotting.
            ax: an axis object to plot to. If none, plt.imshow( ... ) is used.
            bfac (float): a brightness adjustment to apply to RGB mappings (-1 to 1)
            cfac (float): a contrast adjustment to apply to RGB mappings (-1 to 1)
            samples (bool): True if sample points (defined in the header file) should be plotted. Default is False. Otherwise, a list of
                     [ (x,y), ... ] points can be passed.
            tscale (bool): True if each band (for ternary images) should be scaled independently. Default is False.
                    When using scaling, vmin and vmax can be used to set the clipping percentiles (integers) or
                    (constant) values (float).
            rot (bool): if True, the x and y axis will be flipped (90 degree rotation) before plotting. Default is False.
            flipX (bool): if True, the x axis will be flipped before plotting (after applying rotations).
            flipY (bool): if True, the y axis will be flippe before plotting (after applying rotations).
            **kwds: keywords are passed to matplotlib.imshow( ... ), except for the following:

                 - mask = a 2D boolean mask containing true if pixels should be drawn and false otherwise.
                 - path = a file path to save the image too (at matching resolution; use fig.savefig(..) if you want to save the figure).
                 - ticks = True if x- and y- ticks should be plotted. Default is False.
                 - ps, pc = the size and color of sample points to plot. Can be constant or list.
                 - figsize = a figsize for the figure to create (if ax is None).

        Returns:
            Tuple containing

            - fig: matplotlib figure object
            - ax:  matplotlib axes object. If a colorbar is created, (band is an integer or a float), then this will be stored in ax.cbar.
        """

        #create new axes?
        if ax is None:
            fig, ax = plt.subplots(figsize=kwds.pop('figsize', (18,18*self.ydim()/self.xdim()) ))

        # deal with ticks
        if not kwds.pop('ticks', False ):
            ax.set_xticks([])
            ax.set_yticks([])

        #map individual band using colourmap
        if isinstance(band, str) or isinstance(band, int) or isinstance(band, float):
            #get band
            if isinstance(band, str):
                data = self.data[:, :, self.get_band_index(band)]
            else:
                data = self.data[:, :, self.get_band_index(np.abs(band))]
            if not isinstance(band, str) and band < 0:
                data = np.nanmax(data) - data # flip

            # convert integer vmin and vmax values to percentiles
            if 'vmin' in kwds:
                if isinstance(kwds['vmin'], int):
                    kwds['vmin'] = np.nanpercentile( data, kwds['vmin'] )
            if 'vmax' in kwds:
                if isinstance(kwds['vmax'], int):
                    kwds['vmax'] = np.nanpercentile( data, kwds['vmax'] )

            #mask nans (and apply custom mask)
            mask = np.isnan(data)
            if not np.isnan(self.header.get_data_ignore_value()):
                mask = mask + data == self.header.get_data_ignore_value()
            if 'mask' in kwds:
                mask = mask + kwds.get('mask')
                del kwds['mask']
            data = np.ma.array(data, mask = mask > 0 )

            # apply rotations and flipping
            if rot:
                data = data.T
            if flipX:
                data = data[::-1, :]
            if flipY:
                data = data[:, ::-1]

            # save?
            if 'path' in kwds:
                path = kwds.pop('path')
                from matplotlib.pyplot import imsave
                if not os.path.exists(os.path.dirname(path)):
                    os.makedirs(os.path.dirname(path)) # ensure output directory exists
                imsave(path, data.T, **kwds)  # save the image

            ax.cbar = ax.imshow(data.T, interpolation=kwds.pop('interpolation', 'none'), **kwds) # change default interpolation to None

        #map 3 bands to RGB
        elif isinstance(band, tuple) or isinstance(band, list):
            #get band indices and range
            rgb = []
            for b in band:
                if isinstance(b, str):
                    rgb.append(self.get_band_index(b))
                else:
                    rgb.append(self.get_band_index(np.abs(b)))

            #slice image (as copy) and map to 0 - 1
            img = np.array(self.data[:, :, rgb]).copy()
            if np.isnan(img).all():
                print("Warning - image contains no data.")
                return ax.get_figure(), ax

            # invert if needed
            for i,b in enumerate(band):
                if not isinstance(b, str) and (b < 0):
                    img[..., i] = np.nanmax(img[..., i]) - img[..., i]

            # do scaling
            if tscale: # scale bands independently
                for b in range(3):
                    mn = kwds.get("vmin", float(np.nanmin(img)))
                    mx = kwds.get("vmax", float(np.nanmax(img)))
                    if isinstance (mn, int):
                        assert mn >= 0 and mn <= 100, "Error - integer vmin values must be a percentile."
                        mn = float(np.nanpercentile(img[...,b], mn ))
                    if isinstance (mx, int):
                        assert mx >= 0 and mx <= 100, "Error - integer vmax values must be a percentile."
                        mx = float(np.nanpercentile(img[...,b], mx ))
                    img[...,b] = (img[..., b] - mn) / (mx - mn)
            else: # scale bands together
                mn = kwds.get("vmin", float(np.nanmin(img)))
                mx = kwds.get("vmax", float(np.nanmax(img)))
                if isinstance(mn, int):
                    assert mn >= 0 and mn <= 100, "Error - integer vmin values must be a percentile."
                    mn = float(np.nanpercentile(img, mn))
                if isinstance(mx, int):
                    assert mx >= 0 and mx <= 100, "Error - integer vmax values must be a percentile."
                    mx = float(np.nanpercentile(img, mx))
                img = (img - mn) / (mx - mn)

            #apply brightness/contrast mapping
            img = np.clip((1.0 + cfac) * img + bfac, 0, 1.0 )

            #apply masking so background is white
            img[np.logical_not( np.isfinite( img ) )] = 1.0
            if 'mask' in kwds:
                img[kwds.pop("mask"),:] = 1.0

            # apply rotations and flipping
            if rot:
                img = np.transpose( img, (1,0,2) )
            if flipX:
                img = img[::-1, :, :]
            if flipY:
                img = img[:, ::-1, :]

            # save?
            if 'path' in kwds:
                path = kwds.pop('path')
                from matplotlib.pyplot import imsave
                if not os.path.exists(os.path.dirname(path)):
                    os.makedirs(os.path.dirname(path)) # ensure output directory exists
                imsave(path, np.transpose( np.clip( img*255, 0, 255).astype(np.uint8), (1, 0, 2)))  # save the image

            # plot samples?
            ps = kwds.pop('ps', 5)
            pc = kwds.pop('pc', 'r')
            if samples:
                if isinstance(samples, list) or isinstance(samples, np.ndarray):
                    ax.scatter([s[0] for s in samples], [s[1] for s in samples], s=ps, c=pc)
                else:
                    for n in self.header.get_class_names():
                        points = np.array(self.header.get_sample_points(n))
                        ax.scatter(points[:, 0], points[:, 1], s=ps, c=pc)

            #plot
            ax.imshow(np.transpose(img, (1,0,2)), interpolation=kwds.pop('interpolation', 'none'), **kwds)
            ax.cbar = None  # no colorbar

        return ax.get_figure(), ax

    def createGIF(self, path, bands=None, figsize=(10,10), fps=10, **kwds):
        """
        Create and save an animated gif that loops through the bands of the image.

        Args:
            path (str): the path to save the .gif
            bands (tuple): Tuple containing the range of band indices to draw. Default is the whole range.
            figsize (tuple): the size of the image to draw. Default is (10,10).
            fps (int): the framerate (frames per second) of the gif. Default is 10.
            **kwds: keywords are passed directly to matplotlib.imshow. Use this to specify cmap etc.
        """

        frames = []
        if bands is None:
            bands = (0,self.band_count())
        else:
            assert 0 < bands[0] < self.band_count(), "Error - invalid range."
            assert 0 < bands[1] < self.band_count(), "Error - invalid range."
            assert bands[1] > bands[0], "Error - invalid range."

        #plot frames
        for i in range(bands[0],bands[1]):
            fig, ax = plt.subplots(figsize=figsize)
            ax.imshow(self.data[:, :, i], **kwds)
            fig.canvas.draw()
            frames.append(np.frombuffer(fig.canvas.tostring_rgb(), dtype='uint8'))
            frames[-1] = np.reshape(frames[-1], (fig.canvas.get_width_height()[1], fig.canvas.get_width_height()[0], 3))
            plt.close(fig)

        #save gif
        imageio.mimsave( os.path.splitext(path)[0] + ".gif", frames, fps=fps)

    ## masking
    def mask(self, mask=None, flag=np.nan, invert=False, crop=False, bands=None):
        """
         Apply a mask to an image, flagging masked pixels with the specified value. Note that this applies the mask to the
         image in-situ.

         Args:
            flag (float): the value to use for masked pixels. Default is np.nan
            mask (ndarray): a numpy array defining the mask polygon of the format [[x1,y1],[x2,y2],...]. If None is passed then
                    pickPolygon( ... ) is used to interactively define a polygon. If a file path is passed then the polygon
                    will be loaded using np.load( ... ). Alternatively if mask.shape == image.shape[0,1] then it is treated as a
                    binary image mask (must be boolean) and True values will be masked across all bands. Default is None.
            invert (bool): if True, pixels within the polygon will be masked. If False, pixels outside the polygon are masked. Default is False.
            crop (bool): True if rows/columns containing only zeros should be removed. Default is False.
            bands (tuple): the bands of the image to plot if no mask is specified. If None, the middle band is used.

         Returns:
            Tuple containing

            - mask (ndarray): a boolean array with True where pixels are masked and False elsewhere.
            - poly (ndarray): the mask polygon array in the format described above. Useful if the polygon was interactively defined.
         """

        if mask is None:  # pick mask interactively
            if bands is None:
                bands = int(self.band_count() / 2)

            regions = self.pickPolygons(region_names=["mask"], bands=bands)

            # the user bailed without picking a mask?
            if len(regions) == 0:
                print("Warning - no mask picked/applied.")
                return

            # extract polygon mask
            mask = regions[0]

        # convert polygon mask to binary mask
        if mask.shape[1] == 2:

            # build meshgrid with pixel coords
            xx, yy = np.meshgrid(np.arange(self.xdim()), np.arange(self.ydim()))
            xx = xx.flatten()
            yy = yy.flatten()
            points = np.vstack([xx, yy]).T  # coordinates of each pixel

            # calculate per-pixel mask
            mask = path.Path(mask).contains_points(points)
            mask = mask.reshape((self.ydim(), self.xdim())).T

            # flip as we want to mask (==True) outside points (unless invert is true)
            if not invert:
                mask = np.logical_not(mask)

        # apply binary image mask
        assert mask.shape[0] == self.data.shape[0] and mask.shape[1] == self.data.shape[1], \
            "Error - mask shape %s does not match image shape %s" % (mask.shape, self.data.shape)
        for b in range(self.band_count()):
            self.data[:, :, b][mask] = flag

        # crop image
        if crop:
            # calculate non-masked pixels
            valid = np.logical_not(mask)

            # integrate along axes
            xdata = np.sum(valid, axis=1) > 0.0
            ydata = np.sum(valid, axis=0) > 0.0

            # calculate domain containing valid pixels
            xmin = np.argmax(xdata)
            xmax = xdata.shape[0] - np.argmax(xdata[::-1])
            ymin = np.argmax(ydata)
            ymax = ydata.shape[0] - np.argmax(ydata[::-1])

            # crop
            self.data = self.data[xmin:xmax, ymin:ymax, :]

        return mask

    def crop_to_data(self):
        """
        Remove padding of nan or zero pixels from image. Note that this is performed in place.
        """

        valid = np.isfinite(self.data).any(axis=-1) & (self.data != 0).any(axis=-1)
        ymin, ymax = np.percentile(np.argwhere(np.sum(valid, axis=0) != 0), (0, 100))
        xmin, xmax = np.percentile(np.argwhere(np.sum(valid, axis=1) != 0), (0, 100))
        self.data = self.data[int(xmin):int(xmax), int(ymin):int(ymax), :]  # do clipping

    ##################################################
    ## Interactive tools for picking regions/pixels
    ##################################################
    def pickPolygons(self, region_names, bands=0):
        """
        Creates a matplotlib gui for selecting polygon regions in an image.

        Args:
            names (list, str): a list containing the names of the regions to pick. If a string is passed only one name is used.
            bands (tuple): the bands of the image to plot.
        """

        if isinstance(region_names, str):
            region_names = [region_names]

        assert isinstance(region_names, list), "Error - names must be a list or a string."

        # set matplotlib backend
        backend = matplotlib.get_backend()
        matplotlib.use('Qt5Agg')  # need this backend for ROIPoly to work

        # plot image and extract roi's
        fig, ax = self.quick_plot(bands)
        roi = MultiRoi(roi_names=region_names)
        plt.close(fig)  # close figure

        # extract regions
        regions = []
        for name, r in roi.rois.items():
            # store region
            x = r.x
            y = r.y
            regions.append(np.vstack([x, y]).T)

        # restore matplotlib backend (if possible)
        try:
            matplotlib.use(backend)
        except:
            print("Warning: could not reset matplotlib backend. Plots will remain interactive...")
            pass

        return regions

    def pickPoints(self, n=-1, bands=hylite.RGB, integer=True, title="Pick Points", **kwds):
        """
        Creates a matplotlib gui for picking pixels from an image.

        Args:
            n (int): the number of pixels to pick, or -1 if the user can select as many as they wish. Default is -1.
            bands (tuple): the bands of the image to plot. Default is HyImage.RGB
            integer (bool): True if points coordinates should be cast to integers (for use as indices). Default is True.
            title (str): The title of the point picking window.
            **kwds: Keywords are passed to HyImage.quick_plot( ... ).

        Returns:
            A list containing the picked point coordinates [ (x1,y1), (x2,y2), ... ].
        """

        # set matplotlib backend
        backend = matplotlib.get_backend()
        matplotlib.use('Qt5Agg')  # need this backend for ROIPoly to work

        # create figure
        fig, ax = self.quick_plot( bands, **kwds )
        ax.set_title(title)

        # get points
        points = fig.ginput( n )

        if integer:
            points = [ (int(p[0]), int(p[1])) for p in points ]

        # restore matplotlib backend (if possible)
        try:
            matplotlib.use(backend)
        except:
            print("Warning: could not reset matplotlib backend. Plots will remain interactive...")
            pass

        return points

    def pickSamples(self, names=None, store=True, **kwds):
        """
        Pick sample probe points and store these in the image header file.

        Args:
            names (str, list): the name of the sample to pick, or a list of names to pick multiple.
            store (bool): True if sample should be stored in the image header file (for later access). Default is True.
            **kwds: Keywords are passed to HyImage.quick_plot( ... )

        Returns:
            a list containing a list of points for each sample.
        """

        if isinstance(names, str):
            names = [names]

        # pick points
        points = []
        for s in names:
            pnts = self.pickPoints(title="%s" % s, **kwds)
            if store:
                self.header['sample %s' % s] = pnts # store in header
            points.append(pnts)
        # add class to header file
        if store:
            cls_names = self.header.get_class_names()
            if cls_names is None:
                cls_names = []
            self.header['class names'] = cls_names + names

        return points








import unittest
import numpy as np
from tempfile import mkdtemp
import shutil
import os
from hylite.project import proj_persp
from hylite.project import PMap
from hylite import io

class MyTestCase(unittest.TestCase):
    def test_projection(self):

        # generate a basic geometry
        x, y = np.meshgrid( np.linspace(-10,10), np.linspace(-10,10) )
        xyz = np.vstack( [x.ravel(),y.ravel(),np.zeros_like(x.ravel())] ).T

        # project onto a camera
        img_dims = (20,20)
        pp, vis = proj_persp(xyz, np.array([0, 0, 40]),
                             a=np.array([0, 0, 90]), fov=25., dims=img_dims)

        # build projection map
        pm = PMap(*img_dims, xyz.shape[0])
        pm.set_ppc(pp, vis)

        # test basic stuff
        self.assertEqual(pm.point_count(), 1936)
        self.assertEqual(pm.pixel_count(), 400)
        self.assertEqual(pm.get_pixel_indices( 505 )[0][0,1], 18 )
        self.assertAlmostEqual(pm.get_depth( (16, 18), 505 ), 40., places=5 )
        self.assertEqual(pm.get_point_index( (16,18) )[0], 505 )
        self.assertEqual(pm.get_pixel_index( 505 )[0], (16,18), 40.)

        # test points per pixel
        npoints = pm.points_per_pixel()
        self.assertEqual(np.max(npoints.data), 9)

        npixels = pm.pixels_per_point()
        self.assertEqual(np.max(npixels), 1)

    def test_intersections(self):
        # generate a basic geometry
        x, y = np.meshgrid(np.linspace(-10, 10), np.linspace(-10, 10))
        xyz = np.vstack([x.ravel(), y.ravel(), np.zeros_like(x.ravel())]).T

        # project onto a camera 1 and build pmap
        img_dims = (40, 20)
        pp, vis1 = proj_persp(xyz, np.array([-5, 0, 40]),
                              a=np.array([0, 0, 90]), fov=25., dims=img_dims)
        pm1 = PMap(*img_dims, xyz.shape[0])
        pm1.set_ppc(pp, vis1)

        # plt.scatter( pp[vis1,0], pp[vis1,1] )
        # plt.show()

        # project onto a camera 2 and build pmap
        img_dims = (20, 20)
        pp, vis2 = proj_persp(xyz, np.array([5, 0, 40]),
                              a=np.array([0, 0, 90]), fov=25., dims=img_dims)
        pm2 = PMap(*img_dims, xyz.shape[0])
        pm2.set_ppc(pp, vis2)

        U = np.sum(vis1 + vis2)  # number of points in union
        X = np.sum(vis1 & vis2)  # number of points in intersection

        self.assertEqual(len(pm1.union(pm2)), U)
        self.assertEqual(len(pm1.intersect(pm2)), X)

        kk, jj = pm1.intersect_pixels( pm2 )
        self.assertEqual(len(kk), X)
        self.assertEqual(len(jj), X)

    def test_io(self):

        # generate a basic geometry
        x, y = np.meshgrid(np.linspace(-10, 10), np.linspace(-10, 10))
        xyz = np.vstack([x.ravel(), y.ravel(), np.zeros_like(x.ravel())]).T

        # project onto a camera
        img_dims = (20, 20)
        pp, vis = proj_persp(xyz, np.array([0, 0, 40]),
                             a=np.array([0, 0, 90]), fov=25., dims=img_dims)

        # build projection map
        pm = PMap(*img_dims, xyz.shape[0])
        pm.set_ppc(pp, vis)

        pth = mkdtemp()

        # create temp directory
        try:
            # save pmap
            path = os.path.join(pth,"pmap.prj.npz")
            io.save(path, pm)

            # load pmap
            pm2 = io.load( path )

        except:
            shutil.rmtree(pth)  # delete temp directory
            self.assertFalse(True, "Error - could not save PMap data" )

        # checks
        self.assertTrue(os.path.exists(path), "Error - file could not be written.")
        self.assertEqual(pm.point_count(), pm2.point_count())
        self.assertEqual(pm.pixel_count(), pm2.pixel_count())

        shutil.rmtree(pth)  # delete temp directory

if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
from tests import genImage
import hylite

class TestHyImage(unittest.TestCase):
    def test_image(self):

        # test constructor
        image = hylite.HyImage(np.zeros((25,25,5)), wav=np.arange(5)*100)
        self.assertListEqual(list(image.get_wavelengths()), list(np.arange(5)*100))

        # create test image
        image = genImage(dimx = 1464, dimy=401, nbands=10)

        # run plotting functions
        image.quick_plot( (0,1,2), vmin=2, vmax=98 )
        image.quick_plot( 0 )

        self.assertEqual(image.xdim(), 1464)
        self.assertEqual(image.ydim(), 401)
        self.assertEqual(image.band_count(), 10)
        self.assertEqual(image.aspx(),  401 / 1464)

        # todo - add test code for georeferencing code
        # get_extent, set_projection, set_projection_EPSG, get_projection_EPSG, pix_to_world, world_to_pix

        image.flip(axis='y')
        image.data[10,10,:] = np.nan
        image.fill_holes()
        self.assertEqual( np.isfinite( image.data ).all(), True )
        image.blur()

        # resize
        nx,ny = int(1464/2), int(401/2)
        image.resize(newdims=(nx, ny))
        self.assertEqual(image.xdim(), nx)
        self.assertEqual(image.ydim(), ny)
        self.assertEqual(image.band_count(), 10)

        # extract features
        k, d = image.get_keypoints( band=0 )
        src, dst = image.match_keypoints(k,k,d,d)
        self.assertGreater(len(src), 0 ) # make sure there are some matches...

        # masking
        image.mask( np.sum(image.data,axis=2) > 0.75 )
        self.assertEqual(np.isfinite(image.data).all(), False)

if __name__ == '__main__':
    unittest.main()



"""
Simplify file input and output using HyCollection directory structures.
"""

import os
import hylite
import numpy as np
import shutil
import re
from natsort import natsorted
class External(object):
    """
    Small wrapper class for storing external objects in HyCollections.
    """
    def __init__( self, path, base=None ):
        self.path = path
        self.base = base
        self.value = None

    def get(self):

        # is attr already loaded?
        if self.value is not None:
            return self.value

        # no - figure out where it is
        if self.base is not None:
            path = os.path.join(self.base, self.path)
        assert os.path.exists(path), "Error: %s does not exist." % path

        # and load it
        try:
            self.value = hylite.io.load( path )
            return self.value
        except:
            assert False, "Error loading external attribute %s" % path

class HyCollection(object):
    """
    A utility class for mapping data stored in a special file system (.hyc directory or similar) between RAM and disk storage.
    Useful for (1) reducing IO code and (2) writing out-of-core analyses. The underlying .hyc directory can contain any
    data that can be read and written by hylite.io, including: numpy arrays, numbers, strings and hylite.HyData instances.
    """

    def __init__(self, name, root, header=None, vb=False):
        """
        Create a new HyCollection.

        Args:
            name (str): a name for this HyCollection. Names of HyCollections stored in any given directory must be
                  be unique to avoid conflicts.
            root (str): the location of this HyCollection on disk.
            header (hylite.HyHeader): a header file for this HyCollection. If None (default) a new header will be created.
            vb (bool): True if print notifications should be written when data is being loaded from disk. Default is False.
        """
        self.name = os.path.splitext(name)[0]  # trim extension just in case
        self.root = root
        if header is None:
            header = hylite.HyHeader()
        header['file type'] = 'Hylite Collection'  # ensure file type is correct (just in case someone cares)
        self.header = header
        self.vb = vb
        self.ext = '.hyc'

    def get_file_dictionary(self, root=None, name=None):
        """
        Convert this object to a dictionary of files (keys) and serializable objects (values).
        Note that primitive attributes (string, integer, etc.) will be stored in the header file.

        Args:
            root (str): the directory to store this HyCollection in. Defaults to the root directory specified when
                  this HyCollection was initialised, but this can be overriden for e.g. saving in a new location.
            name (str): the name to use for the HyCollection in the file dictionary. If None (default) then this instance's
                  name will be used, but this can be overriden for e.g. saving in a new location.
        Returns:
            a dictionary such that dict[ path ] = object.
        """
        # parse root
        if root is None:
            root = self.root
        if name is None:
            name = self.name
        assert root is not None, "Error - root argument must be set during HyCollection initialisation or function call."
        assert name is not None, "Error - name argument must be set during HyCollection initialisation or function call."

        # get all attributes (excluding class methods/variables and the header variable)
        attr = self.getAttributes()

        # build paths dictionary
        out = {os.path.join(root, "%s.hdr" % name): self.header}
        path = self.getDirectory(root=root, name=name)
        for a in attr:
            value = getattr(self, a)  # get value
            if value is None:
                continue # skip None values
            elif type(value) in [int, str, bool, float]:  # primitive types
                self.header[a] = value  # store in header file
            elif isinstance(value, External): # external link
                self.header[a] = "<" + value.path + ">" # store in header file
            else:
                out[os.path.join(path, a)] = value

        return out

    def clean(self):
        """
        Delete files associated with attributes that have been cleared by setting to None.
        """
        attr = self.getAttributes()
        for a in attr:
            value = getattr(self, a)
            if value is None:
                # remove from header file? (easy)
                if a in self.header:
                    del self.header[a]
                else:
                    # remove from disk...
                    # solve path from attribute name
                    path = None
                    for f in os.listdir(self.getDirectory()):
                        if os.path.splitext(f)[0] == a:  # we have found the right file
                            path = os.path.join(self.getDirectory(), f)
                            break
                    if path is not None and os.path.exists(path):
                        hdr, dat = hylite.io.matchHeader( path )
                        if hdr is not None and os.path.exists(hdr):
                            os.remove(hdr)
                        if dat is not None and os.path.exists(dat) and os.path.isdir(dat): # nested HyCollection
                            shutil.rmtree(dat)
                        if os.path.exists(dat) and os.path.isfile(dat): # other data type
                            os.remove(dat)
                # remove attribute
                delattr(self, a)

    def _loadAttribute_(self, attr):
        """
        Load an attribute from disk. This should generally not be called directly - rather it is called
        if needed when any class attributes are requested.

        Args:
            attr (str): the attribute name to load.
        """

        # check if attribute is in the header file
        attr = attr.strip().replace(' ', '_')

        # edge case - the attribute being requested is the header file!
        if 'header' == attr or '__' in attr:  # ignore headers and private (__x__) variables.
            raise AttributeError(attr)

        # try loading attribute from disk
        try:
            # no file associated with this HyCollection - raise attribute error.
            if not os.path.exists(self.getDirectory(makedirs=False)):
                raise AttributeError

            # solve path from attribute name
            path = None
            if (attr + '.hdr') in os.listdir(self.getDirectory(makedirs=False)):
                # look for header files first
                # (in case e.g. attr.hdr, attr.png and attr.dat all exist)!
                path = os.path.join(self.getDirectory(makedirs=False), attr + '.hdr')
            else:
                # no header file; search for any extension
                for f in os.listdir(self.getDirectory(makedirs=False)):
                    if os.path.splitext(f)[0] == attr:  # we have found the right file
                        path = os.path.join(self.getDirectory(makedirs=False), f)
                        break
            if (path is None) or not os.path.exists(path):
                raise AttributeError # couldn't find on disk

            #assert path is not None and os.path.exists(path), \
            #    "Error - could not load attribute %s from disk (%s)." % ( attr, self.getDirectory(makedirs=False))

            # load attribute
            if self.vb:
                print("Loading %s from %s" % (attr, path))
            self.__setattr__(attr, hylite.io.load(path))  # load and update HyCollection attribute
        except AttributeError:

            # last chance - look in header data
            if (attr in self.header) or (attr.replace('_', ' ') in self.header):
                # get value from header file
                if attr in self.header:
                    val = self.header[attr]  # no spaces
                else:
                    val = self.header[attr.replace('_', ' ')]  # possibly has spaces

                # parse strings if needed
                if isinstance(val, str):
                    # load as a list
                    if '{' in val and '}' in val:
                        val = self.header.get_list(val)
                    elif '<' in val and '>' in val:  # load as an external path
                        val = self.header[val].strip()[1:-1]
                        if os.path.exists(val):  # absolute path!
                            val = External(val, None)  # wrap in absolute External class
                        else:
                            val = External(val, self.root)  # wrap in relative External class
                    else:  # is it an integer or a float?
                        try:
                            val = int(val)
                        except:
                            try:
                                val = float(val)
                            except:
                                # finally, try converting boolean types
                                if val.lower() == 'true':
                                    val = True
                                elif val.lower() == 'false':
                                    val = False

                # done
                self.__setattr__(attr, val)  # easy!
            else:
                raise AttributeError

    def get_path(self, name: str):
        """
        Return the path of the specified attribute. Note that this file may or may not exist, depending
        on if this HyCollection has been saved previously. Also note that this path will exclude the file extension.
        """
        if name == 'header' or name in self.header:
            return os.path.splitext( self.getDirectory() )[0]
        else:
            name = name.strip().replace(' ', '_')
            return os.path.join( self.getDirectory(), name )

    def getDirectory(self, root=None, name=None, makedirs=False):
        """
        Return the directory files associated with the HyCollection are stored in.

        Args:
            root (str): the directory to store this HyCollection in. Defaults to the root directory specified when
                  this HyCollection was initialised, but this can be overriden for e.g. saving in a new location.
            name (str): the name to use for the HyCollection in the file dictionary. If None (default) then this instance's
                  name will be used, but this can be overriden for e.g. saving in a new location.
            makedirs (bool): True if this directory should be created if it doesn't exist. Default is False.
        """
        if root is None:
            root = self.root
        if name is None:
            name = self.name
        assert root is not None, "Error - root argument must be set during HyCollection initialisation or function call."
        assert name is not None, "Error - name argument must be set during HyCollection initialisation or function call."
        p = os.path.join(root, os.path.splitext(name)[0] + self.ext)
        if makedirs:
            os.makedirs(p, exist_ok=True)  # ensure directory actually exists!
        return p

    def getAttributes(self, ram_only=True, file_formats=False):
        """
        Return a list of available attributes in this HyCollection.

        Args:
            ram = True if only attributes loaded in RAM should be included. Default is True.
            file_formats = True if the file extensions of attributes stored on disk should be retained. Default is False.
        """
        # get potential attributes
        attr = list(set(dir(self)) - set(dir(HyCollection)) - set(['header', 'root', 'file type','name', 'ext', 'vb']))

        # loop through and remove all functions
        out = []
        known = []
        for a in attr:
            if '__' in a:
                continue # ignore private variables
            if not type(super().__getattribute__(a)) == type(self.getAttributes): # ignore methods
                if file_formats:
                    out.append((a,type(self.get(a)).__name__))
                else:
                    out.append(a)
                known.append(a)

        # add elements in header
        for a,v in self.header.items():
            if a not in known:
                if file_formats:
                    if file_formats:
                        out.append((a, type(self.get(a)).__name__))
                    else:
                        out.append(a)
                    known.append(a)

        # also add attributes on disk
        if not ram_only:
            if os.path.exists(self.getDirectory()):
                for f in os.listdir(self.getDirectory()):
                    name, ext = os.path.splitext(f)
                    if name not in known and ext != '.hdr' and ext != '':
                        if file_formats:
                            out.append((name,ext))
                        else:
                            out.append(name)
                        known.append(name)
        return out

    def query(self, *, name_pattern=None, ext_pattern=None, recurse=False, recurse_matches=False, ram_only=False):
        """
        Finds attributes of this HyCollection with names or types matching the specified patterns. Note that (1)
        if both name_pattern and ext_pattern are provided then attributes must match both filters to be included in the
        results, and (2)

        Args:
            name_pattern (list, str) = A regex pattern (string) or list of regex pattern strings to match against. If an attribute name
                          matches against any of the provided patterns then it will be included in the output.
                          Default is None (match all attributes).
            ext_pattern (list, str) = A regex pattern (string) or list of regex pattern strings to match against. Matches will be evaluated
                         against file extensions for attributes on the disk (e.g., ".hdr") and type names (e.g., "HyImage") for
                         attributes loaded in RAM (as we cannot guess what their file extension may be). Note that class inheritance
                         is not considered during this matching, so e.g., "HyData" will not match with "HyImage".
                         Default is None (match all attributes).
            recurse (bool) = True if (all) child HyCollections should also be queried to search the entire HyCollection tree for
                     matches. Default is False.
            recurse_matches (bool) = True if HyCollections that match the provided filters should also be queried recursively. Default
                    is False.
            ram_only (bool) = True if only attributes already loaded into memory should be queried. Default is False.
        """

        def match(value, pattern):
            """
            Quick function for matching against list or regex.
            """
            if pattern is None:
                return True
            if not isinstance(pattern, list) or isinstance(pattern, tuple) or isinstance(pattern, np.ndarray):
                pattern = [pattern]
            for p in pattern:
                assert isinstance(p,str), "Error - %s is an invalid pattern [ should be string ]" % p
                if re.search(p, value) is not None:
                    return True  # we have a match!
            return False

        attr = self.getAttributes(ram_only=ram_only, file_formats=True)
        out = []
        for a,e in attr:

            # do matching
            if match(a, name_pattern) and match(e, ext_pattern):
                out.append(a)
                if not recurse_matches:
                    continue # skip to next loop

            # recurse if required
            if recurse:
                if self.loaded(a): # object already loaded in ram
                    if isinstance( self.get(a), hylite.HyCollection ):
                        out += self.get(a).query( name_pattern=name_pattern, ext_pattern=ext_pattern,
                                                  recurse=recurse, recurse_matches=recurse_matches,
                                                  ram_only = ram_only )
                else: # object is on the disk
                    from hylite.io import _loadCollection
                    try:
                        C = _loadCollection(os.path.join(self.getDirectory(), a + e) )
                        out += C.query( name_pattern=name_pattern, ext_pattern=ext_pattern,
                                                      recurse=recurse, recurse_matches=recurse_matches,
                                                      ram_only = ram_only )
                    except:
                        pass # continue, this was not a HyCollection
        return natsorted(out) # sort alphabetically for consistency

    def loaded(self, name):
        """
        Return True if the requested attribute is loaded into RAM already, and False if it exists on the disk.
        Throws an attribute error if the attribute does not exist.
        """
        try:  # try getting attribute
            attr = object.__getattribute__(self, name.strip().replace(' ','_'))
            return True
        except:
            files = self.getAttributes(ram_only=False)
            assert name in files, "Error - attribute %s does not exist." % name
            return False

    def print(self):
        """
        Print a nicely formatted summary of the contents of this collection.
        """
        attr = self.getAttributes()

        # print loaded variables
        print("Attributes stored in RAM:")
        for a in attr:
            v = getattr(self, a)
            print("\t - %s called %s" % (type(v), a) )

        # print header variables
        print("Attributes stored in header:")
        for k,v in self.header.items():
            if k not in ['file type','file_type', 'path'] and k not in attr: # header keys to ignore
                if isinstance(v,str) or isinstance(v, list) or isinstance(v, np.ndarray):
                    print("\t %s = %s"% (k,type(v))) # print type
                else:
                    print("\t %s = %s" % (k, v)) # print value

        # print disk variables
        if os.path.exists(self.getDirectory()):
            print("Attributes stored on disk:")
            for f in os.listdir(self.getDirectory()):
                name, ext = os.path.splitext(f)
                if name not in attr and ext != '.hdr':
                    print( "\t - %s" % f)

    def save(self):
        """
        Quick utility function for saving this in the predefined location.
        """
        from hylite import io # occasionally io doesn't seem to get loaded unless we call this ... strange?
        hylite.io.save(os.path.splitext(self.getDirectory())[0], self)

    def save_attr(self, attr):
        """
        Save a single attribute in this HyCollection.
        """
        from hylite import io  # occasionally io doesn't seem to get loaded unless we call this ... strange?
        if attr in self.header:
            hylite.io.saveHeader( os.path.splitext(self.getDirectory())[0] + '.hdr', self.header )
        else:
            hylite.io.save( self.get_path(attr), self.get(attr) )

    def free(self):
        """
        Free all attributes in RAM. To avoid losing data, be sure to save this HyCollection first (e.g. using
        self.save(...).
        """
        attr = self.getAttributes()
        for a in attr:
            delattr(self, a)

    def free_attr(self, a):
        """
        Unload the specific attribute from RAM. Can be used to e.g. remove unchanged attributes before saving. Note that any unsaved
        changes to this attribute will be lost.
        """
        delattr(self, a)

    def addExternal(self, name, path, relative=True):
        """
        Add an external link (that is saved/loaded by this HyCollection instance, but not stored in its data folder).

        Args:
            name: the name of the attribute to add.
            path: the path to the object to add.
            relative: True if the path should be converted to a relative one. Default is True.
        """

        assert os.path.exists(path), "Error - %s is not a valid file or folder." % path

        if relative:
            path = os.path.relpath( path, self.root )
            self.__setattr__(name, External(path,self.root))
        else:
            self.__setattr__(name, External(path, None))

    def addSub(self, name):
        """
        Add a subcollection to this one (and sort out internal file paths etc.).

        Args:
            name (str): the name of the subcollection to add.

        Returns:
            a HyCollection object representing the subcollection.
        """

        if self.root is None: # no path defined
            S = HyCollection(name,'')
        else:
            S = HyCollection(name, self.getDirectory(makedirs=False))
        self.__setattr__(name, S)
        return S

    # expose getters and setters for easy access
    def get(self, name):
        """
        Get an attribute in this collection.
        """
        return self.__getattribute__(name)

    def set(self, name, value, save=False):
        """
        Set an attribute in this collection.

        Args:
            name = the name of the variable to set.
            value = the value to set this variable too.
            save = True if the variable should immediately be saved to disk.
        """
        self.__setattr__(name, value)
        if save:
            self.save_attr(name)

    def __getattribute__(self, name):
        """
        Override __getattribute__ to automatically load out-of-core attributes if they are asked for.
        """
        # attribute has not yet been loaded - do so now
        try:  # try getting attribute
            attr = object.__getattribute__(self, name.strip().replace(' ', '_'))
        except AttributeError:  # no attribute found
            self._loadAttribute_(name)  # load the attribute from disk
            if not hasattr(self, name):
                raise AttributeError(name)
            else:
                attr = object.__getattribute__(self, name.strip().replace(' ', '_') )

        # resolve external links if necessary
        if isinstance(attr, External):
            return attr.get()

        # return
        return attr

    def __setattr__(self, name, value):
        """
        Override __setattr__ to throw an error when dodgy data types are added to this collection. Note that private
        attributes (with __ in their name, such as __foo__) are allowed to have any type, but will not be written to
        disk.
        """
        if '__' not in name:
            valid = type(value) in [ int, str, bool, float, list ] # accept primitive types
            valid = valid or isinstance( value, np.ndarray) # accept numpy arrays
            valid = valid or isinstance( value, hylite.HyData ) # accept hydata types
            valid = valid or isinstance(value, hylite.HyHeader)  # accept hydata types
            valid = valid or isinstance( value, hylite.HyCollection ) # accept HyCollection instances (nesting)
            valid = valid or isinstance(value, hylite.project.Camera ) # accept Camera instances
            valid = valid or isinstance(value, hylite.project.Pushbroom)  # accept Pushbroom instances
            valid = valid or isinstance(value, hylite.project.PMap)  # accept Pushbroom instances
            valid = valid or isinstance(value, External ) # accept external links
            valid = valid or isinstance(value, list) and np.array( [isinstance(d, hylite.HyData) ] for d in value ).all() # multimwl maps
            valid = valid or value is None # also accept None
            assert valid, "Error - %s is an invalid attribute type for HyCollection." % type(value)

        object.__setattr__(self, name.strip().replace(' ', '_'), value)


"""
Static classes for sensor-specific processing such as lens-correction, adjustments for sensor shift and conversions
from digital numbers to radiance.
"""
import matplotlib.pyplot as plt
from .sensor import Sensor
from .fx import FX10
from .fx import FX17
from .fx import FX50
from .owl import OWL
from .fenix import Fenix
from .rikola import Rikola, Rikola_HSC2, Rikola_RSC1
from .fx import *

# noinspection PyDefaultArgument
def QAQC(image, method, dim=0, fit="minmax", checklines=[]):
    """
    Estimate the spectral quality of a sensor according to reference measurements. Mask image first if required.

    Args:
        image (hylite.HyImage): the image containing data from the sensor..
        method (str): "LDPE" for SWIR using reference LDPE foil, "FT" for VNIR using fluorescence tube.
        dim (int): dimensionality of the evaluation (0 = overall average, 1 = row-wise, 2 = full frame).
        fit (str): method for peak fitting. For details, check hylite.analyse.mapping.minimum_wavelength( ... ).
        checklines (list): define custom features to check for (list of ints or floats).
    """

    image.data = image.data.astype(np.float32)

    # define indicative feature lines
    if method == "FT":
        checklines = [404.66, 435.83, 546.08, 611.08]
        image.data = np.nanmax(image.data) - image.data
    elif method == "LDPE":
        checklines = [1728., 1764., 2310., 2350.]

    # calculate and plot accuracy assessment depending on defined dimensionality
    if dim == 0:
        image.data = np.mean(image.data, axis=(0, 1))
        for line in checklines:
            lim = image.get_fwhm()[image.get_band_index(float(line))]
            mini, _, _ = image.minimum_wavelength(line - 10., line + 10., method=fit)
            # plot warning if error exceeds FWHM
            if (line - mini) > lim:
                print("\x1b[31m" + "Spectral accuracy at " + str(line) + " nm:   " + "{:.4f}".format(
                    line - mini) + " nm - WARNING: OVER FWHM" + '\x1b[0m')
            else:
                print("\x1b[32m" + "Spectral accuracy at " + str(line) + " nm:   " + "{:.4f}".format(
                    line - mini) + " nm" + '\x1b[0m')

    elif dim == 1:
        image.data = np.mean(image.data, axis=1)
        fig, axs = plt.subplots(2, 2, figsize=(15, 6))
        fig.subplots_adjust(hspace=.5, wspace=.3)
        axs = axs.ravel()
        i = 0
        for line in checklines:
            mini, _, _ = image.minimum_wavelength(line - 10., line + 10., method=fit)

            # color plot based on fwhm
            lim = image.get_fwhm()[image.get_band_index(float(line))]
            if any(x > lim for x in (line - mini)):
                col = "orangered"
            else:
                col = "limegreen"
            axs[i].plot(line - mini, col)
            axs[i].set_title(str(line) + ' nm')
            axs[i].axhline(y=lim, color="grey", linestyle="--")
            axs[i].text(0.9, 0.9, 'FWHM', fontsize=10, color="grey", va='center', ha='center', backgroundcolor='w',
                        transform=axs[i].transAxes)
            i += 1
        for ax in axs.flat:
            ax.set(xlabel='swath pixels', ylabel='spectral accuracy [nm]')

        plt.show()

    elif dim == 2:
        from matplotlib import colors
        # color plot based on fwhm
        fig, axs = plt.subplots(2, 2, figsize=(15, 6))
        fig.subplots_adjust(hspace=.5, wspace=.1)
        axs = axs.ravel()
        i = 0
        for line in checklines:
            mini, _, _ = image.minimum_wavelength(line - 10., line + 10., method=fit)
            lim = image.get_fwhm()[image.get_band_index(float(line))]
            cmap = "RdYlGn_r"
            norm = colors.Normalize(vmin=0, vmax=lim * 2)
            im = axs[i].imshow(line - mini, cmap=cmap, norm=norm)
            axs[i].set_title(str(line) + ' nm')
            cbar = fig.colorbar(im, ax=axs[i], cmap=cmap, norm=norm)
            cbar.ax.locator_params(nbins=3)
            cbar.ax.set_yticklabels(['within limits', 'FWHM', 'exceeding FWHM'])
            i += 1

        plt.show()


from hylite import HyData, HyImage, HyCloud, HyHeader
import numpy as np
from pathlib import Path
from hylite import io
import os

# functions for creating test datasets
def genHeader():
    return  io.loadHeader( os.path.join( os.path.join(str(Path(__file__).parent.parent), "test_data"), "image.hdr") )

# create test cloud
def genCloud( npoints = 1000, nbands=10, data=True, normals=True, rgb=True):
    header = genHeader()

    xyz = np.random.rand(npoints, 3)
    if rgb:
        rgb = np.random.rand(npoints, 3)
    if normals:
        normals = np.random.rand(npoints, 3)
        normals = normals / np.linalg.norm(normals, axis=1)[:,None]
    if data:
        data = np.random.rand(npoints, nbands)  # point cloud

    cloud = HyCloud( xyz, rgb = rgb, normals=normals, bands=data, header=header )
    cloud.set_wavelengths( np.linspace(500.0,1000.0,nbands) )
    cloud.set_band_names( ["Band %d" % (i+1) for i in range(nbands)])
    cloud.set_fwhm( [1.0 for i in range(nbands)])
    cloud.push_to_header() # push info on dims

    return cloud

def genImage( dimx = 1464, dimy=401, nbands=10 ):
    header = genHeader()
    data = np.random.rand(dimx, dimy, nbands)  # image
    data = HyImage( data, header = header )
    data.set_wavelengths( np.linspace(500.0,1000.0,nbands) )
    data.set_band_names( ["Band %d" % (i+1) for i in range(nbands)])
    data.set_fwhm( [1.0 for i in range(nbands)])
    data.push_to_header() # push info on dims
    return data


"""
Functions for projecting between pushbroom sensors (with known orientation/position from IMU data) and a 3D point cloud.
"""

import hylite
from hylite.project import rasterize, PMap, push_to_cloud
from scipy import spatial
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
import scipy as sp
from scipy.optimize import minimize


class Pushbroom(object):
    """
    An extension of the Camera class to handle pushbroom sensors.

    A note on coordinates: hylite always considers the x-axis of an image to be in the cross-track direction, and
    the y-axis to be in the along-track (movement) direction. This is consistent with most sensor manufacturers, but
    should not be confused with the coordinates used for panoramic sensors (where x = lines at each rotation, y=cross track
    direction). Issues plotting can be resolved using the rot, flipX and flipY arguments for image.quick_plot( ... ).
    """

    def __init__(self, pos, ori, xfov, lfov, dims ):
        """
        Initialise this pushbroom camera instance.

        Args:
            pos: a list containing the position of the sensor at each frame.
            ori: a list containing the orientation of the sensor (roll, pitch, yaw in degrees) at each frame.
            xfov: the across-track field of view of one pixel (e.g. 0.01 degree).
            lfov: the along-track field of view of one pixel (e.g. 0.01 degree).
            dims: the dimensions of the resulting image. Should be (pixels_in_sensor, number_of_frames).
        """

        # store track and fov info
        self.cp = np.array(pos)  # camera position track
        self.co = np.array(ori)  # camera orientation track
        self.xfov = xfov
        self.lfov = lfov
        self.px = 2 * np.tan(0.5 * np.deg2rad(self.xfov))  # pixel pitch (across track)
        self.pl = 2 * np.tan(0.5 * np.deg2rad(self.lfov))  # pixel pitch (along track)

        # calculate and store rotation matrices
        self.R = [spatial.transform.Rotation.from_euler('xyz', [a[0] - 180, a[1], 90 - a[2]], degrees=True) for a in
                  self.co]
        # N.B. 90 degree rotation of yaw transforms to geographic coordinates (000 = North). 180 degree rotation
        # transforms camera direction from pointing up to down.

        # check and store dims
        assert dims[1] == len(pos), "Error - dims has %d lines, but only %d positions provided." % (dims[1], len(pos))
        assert dims[1] == len(ori), "Error - dims has %d lines, but only %d orientations provided." % (
        dims[1], len(pos))
        self.dims = dims

    def fudge(self, t_all, t_known, cp_known, co_known, method='quadratic'):
        """
        Apply a fudge-factor to fit IMU data to sparse points with known positions / orientations (e.g.
        from SfM results using a co-aligned camera). This (1) matches the two position datasets using the
        two time values (these should be floating point values like GPS seconds), (2) calculates the residual
        at these points, (3) interpolates these residuals between the known points and (4) subtracts them from
        the IMU track.

        Args:
            t_all: timestamps associated with the orientation and position data in this pushbroom track.
            t_known: an array of timestamps of shape (n,) associated with the known positions/orientations.
            cp_known: an array of shape (n,3) containing known positions to fudge to.
            co_known: an array of of shape (n,3) containing known orientations to fudge to.
            method: the interpolation method to use, as defined in scipy.interpolate.interp1d. Default is 'quadratic'.

        Returns:
            a fudged copy of this track.
        """

        # check shapes
        assert t_all.shape[0] == self.co.shape[0], "Error - timestamps t_all has %d entries, but %d are needed." % (t_all.shape[0], self.co.shape[0])
        assert t_known.shape[0] == cp_known.shape[0], "Error - timestamp shape (%d) != cp_known shape (%d)." % (t_known.shape[0], cp_known.shape[0])
        assert t_known.shape[0] == co_known.shape[0], "Error - timestamp shape (%d) != co_known shape (%d)." % (t_known.shape[0], co_known.shape[0])

        # get indices of main track that correspond to known frames
        S = [np.argmin(np.abs(t_all - _t)) for _t in t_known]
        if len(np.unique(S)) != len(S):
            print("Warning - duplicate matches are present; tracks probably do not overlap properly?")

        # compute errors
        e_p = cp_known - self.cp[S, :] # error in position
        e_o = co_known - self.co[S, :] # error in orientation

        # apply fudge to remove these (and interpolate fudge factor between known points)
        cp_adj = self.cp + sp.interpolate.interp1d(t_known, e_p, bounds_error=False, axis=0, kind='quadratic', fill_value=0)(t_all)
        co_adj = self.co + sp.interpolate.interp1d(t_known, e_o, bounds_error=False, axis=0, kind='quadratic', fill_value=0)(t_all)

        # return a copy of this track
        return Pushbroom( cp_adj, co_adj, self.xfov, self.lfov, self.dims )

    def rebuild_R(self):
        """
        Rebuild the internal rotation matrix from the self.co arrays.
        """
        self.R = [spatial.transform.Rotation.from_euler('xyz', [a[0] - 180, a[1], 90 - a[2]], degrees=True) for a in
                  self.co]

    def apply_boresight(self, roll, pitch, yaw ):
        """
        Add constant values (boresight) to the roll, pitch and yaw angles.

        Returns: a new Pushbroom instance with the adjusted values.
        """
        # appy boresight
        co_adj = self.co + np.array( [roll, pitch, yaw ] )
        return Pushbroom( self.cp, co_adj, self.xfov, self.lfov, self.dims )

    def get_R(self, i=None):
        """
        Return scipy.spatial.Rotation objects that store camera orientation data.

        Args:
         i: the frame index to get a rotation object for. If None (Default) a list of all Rotations is returned.
        """
        if i is not None:
            return self.R[i]
        else:
            return self.R

    def get_axis(self, axis, i=None):
        """
        Get the local camera x (0), y (1) or z (2) axis vector from the rotation matrices.

        Args:
            axis: 0 (x), 1 (y) or 2 (z).
            i: the frame index to get vector for. If None (Default) a list of all vectors is returned.
        """
        if i is not None:
            assert i >= 0 and i < 3, "Error - i must be 0, 1 or 2."
            return self.R[i].as_matrix()[:, axis]
        else:
            return np.array([_R.as_matrix()[:, axis] for _R in self.R])

    def get_x(self, i=None):
        """
        Get the Camera's along-track (movement) vector.

        Args:
         i: the frame index to get vector for. If None (Default) a list of all vectors is returned.
        """
        return self.get_axis(0, i)

    def get_y(self, i=None):
        """
        Get the Camera's cross-track vector.

        Args:
         i: the frame index to get vector for. If None (Default) a list of all vectors is returned.
        """
        return self.get_axis(1, i)

    def get_z(self, i=None):
        """
        Get the Camera's view vector.

        Args:
         i: the frame index to get vector for. If None (Default) a list of all vectors is returned.
        """
        return self.get_axis(2, i)

    def project_to_frame(self, cloud, i, flip=True):
        """
        Project the point cloud onto an (instantaneous) frame from this pushbroom camera.

        Args:
            cloud: the point cloud to project. Must have points stored in cloud.xyz (HyCloud) or be a (n,3) array.
            i: the camera frame (along track index) to use.
            flip: True if pixel coordinates should be flipped.
        Returns:
            a (3,) array of projected points with coordinates that are:

             - xtrack = the position (in pixels) of the points across track (pixels on the sensor)
             -  ltrack = the position (in pixels) of the points along track. Values between 0 and 1 will pass through the sensor slit.
             -  depth = the depth along the view direction of each point (in meters).
        """

        # project into camera coordinates
        if isinstance(cloud, np.ndarray):
            #xyz = np.dot(cloud - self.cp[i], self.R[i].as_matrix())
            xyz = (cloud - self.cp[i])@(self.R[i].as_matrix())
        else:
            #xyz = np.dot(cloud.xyz - self.cp[i], self.R[i].as_matrix())
            xyz = (cloud.xyz - self.cp[i])@(self.R[i].as_matrix())

        # calculate along-track coordinate (perspective projection perpendicular to flight line)
        ltrack = 0.5 + (
                    xyz[:, 0] / xyz[:, 2]) / self.pl  # N.B. +0.5 moves from pixel-center coords to pixel-edge coords

        # calculate across-track coordinate (perspective projection onto sensor array)
        if flip:
            xtrack = (self.dims[0] / 2) - ((xyz[:, 1] / xyz[:, 2]) / self.px)
        else:
            xtrack = (self.dims[0] / 2) + ((xyz[:, 1] / xyz[:, 2]) / self.px)

        return np.array([xtrack, ltrack, xyz[:, 2]]).T

    def crop(self, start, end, image=None):
        """
        Clip this tracks IMU data to the specified frames. Note that this is not reversable.

        Args:
            start: the start frame.
            end: the end frame.
            image: a HyImage instance to clip also, if provided (default is None). Clipping will
                   be applied to the y-direction.
        """
        assert self.co.shape[0] > end and start > 0, "Error - invalid range %d - %d (data shape is %s)" % (start,end,self.co.shape)
        assert start < end, "Error - start (%d) cannot be after end (%d)." % (start,end)
        self.co = self.co[start:end, :]
        self.cp = self.cp[start:end, :]
        self.R = self.R[start:end]
        self.dims = (self.dims[0], self.co.shape[0] )
        if image is not None:
            image.data = image.data[ :, start:end, : ]

    def plot_waterfall(self, image = None, bands=(0,1,2), flipY=True, flipX=False, rot=True ):
        """
        Plot IMU data and (if provided) the raw image frames.

        Args:
            image: the waterfall image to plot. Default is None (no plot).
            bands: the bands of the waterfall image to plot. Default is (0,1,2).
            flipY: flip the Y axis of the image. Default is True.
            flipX: flip the X axis of the image. Default is False.
            rot: rotate the image by 90 degrees. Default is True.
        Returns:
            fig,ax = the figure and a list of associated axes.
        """
        if image is None:
            fig, ax = plt.subplots(6, 1, figsize=(18, 12))
            ax = [ax[0],ax[0],ax[1],ax[2],ax[3],ax[4],ax[5]]
        else:
            fig, ax = plt.subplots(7, 1, figsize=(18, 8))
            image.quick_plot(bands, ax=ax[0], rot=rot, flipY=flipY, flipX=flipX)
            ax[0].set_aspect('auto')  # override aspect equal, as this is probably wrong anywa
            ax[0].set_title("Waterfall")
            ax[0].set_yticks([])
        # plot IMU data
        for i,(y,l,c) in enumerate(zip(
            [self.cp[:,0], self.cp[:,1], self.cp[:,2], self.co[:,0], self.co[:,1], self.co[:,2]],
            ['X','Y','Z','Roll','Pitch','Yaw'],
            ['r','g','b','r','g','b'])):

            ax[i+1].set_ylabel(l)
            ax[i+1].plot(y,color=c)

        # clean up axes
        for a in ax:
            a.set_xticks(range(0, self.co.shape[0], 250))
            a.set_xlim(0, self.co.shape[0])
            a.grid()
        [ax[i].set_xticklabels([]) for i in range(6)]
        fig.tight_layout()
        return fig, ax

    def plot_curtain(self, scale=10, alpha=0.1, ax=None, ortho=None):
        """
        Create a curtain plot for visualising this camera track.

        Args:
            scale: the scale of the camera basis vectors (curtain). This is in transformed coordinates.
            alpha: the alpha value to use for the curtain. Default is 0.1.
            ax: an external axis to plot too.
            ortho: an orthoimage to use to project to pixel coordinates. Image must have a defined affine transform.
        """

        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(10, 10))
            ax.set_aspect('equal')

        # get camera basis
        X = self.get_x()
        Y = self.get_y()
        Z = self.get_z()

        # build curtain
        S = []  # store camera pos for plotting
        XV = []
        YV = []
        ZV = []

        for i in range(len(self.cp)):

            # get camera position at this frame
            px = self.cp[i, 0]
            py = self.cp[i, 1]

            if ortho is not None:
                assert ortho.affine is not None, "Error - image must have an affine transform set."
                px, py = ortho.world_to_pix(px, py)

            S.append([px, py])

            if ortho is None:
                ZV += [[px, px + Z[i, 0] * scale], [py, py + Z[i, 1] * scale]]
                YV += [[px, px + Y[i, 0] * scale], [py, py + Y[i, 1] * scale]]
                XV += [[px, px + X[i, 0] * scale], [py, py + X[i, 1] * scale]]
            else:
                # we need to flip the y-axis as imshow puts the origin in the top-left.
                ZV += [[px, px + Z[i, 0] * scale], [py, py - Z[i, 1] * scale]]
                YV += [[px, px + Y[i, 0] * scale], [py, py - Y[i, 1] * scale]]
                XV += [[px, px + X[i, 0] * scale], [py, py - X[i, 1] * scale]]

        ax.plot(*ZV, color='r', lw=1, alpha=alpha, zorder=3)
        ax.plot(*YV, color='g', lw=1, alpha=alpha, zorder=2)
        ax.plot(*XV, color='b', lw=1, alpha=alpha, zorder=1)

        # plot camera track
        S = np.array(S)
        ax.cmap = ax.scatter(S[:, 0], S[:, 1], c=np.arange(0, S.shape[0]), cmap='jet', zorder=4)  # also store colorbar
        return ax.get_figure(), ax

    def plot_pose(self, i):
        """
        Plot the camera axes and scan line in world coordinates. Useful for checking camera orientation at a
        given frame (and debugging!).

        Args:
            i: plot the i'th camera.
        Returns:
            fig,ax = the matplotlib plot.
        """
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import axes3d
        from mpl_toolkits.mplot3d.art3d import Poly3DCollection

        # get local axes from rotation matrix
        M = self.get_R(i).as_matrix()
        x = M[:, 0]
        y = M[:, 1]
        z = M[:, 2]

        # plot them in 3D
        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca(projection='3d')

        ax.quiver([0], [0], [0],
                  [x[0]], [x[1]], [x[2]],
                  length=1.0, normalize=False, colors='b', label='x (flight direction)')
        ax.quiver([0], [0], [0],
                  [y[0]], [y[1]], [y[2]],
                  length=1.0, normalize=False, colors='g', label='y (pixel direction)')
        ax.quiver([0], [0], [0],
                  [z[0]], [z[1]], [z[2]],
                  length=1.0, normalize=False, colors='r', label='z (view direction)')

        # add plot showing scan line
        V0 = np.array([0, 0, 0])
        V1 = z - 0.4 * y
        V2 = z + 0.4 * y
        xx = [V0[0], V1[0], V2[0], V0[0]]
        yy = [V0[1], V1[1], V2[1], V0[1]]
        zz = [V0[2], V1[2], V2[2], V0[2]]
        verts = [list(zip(xx, yy, zz))]
        ax.add_collection3d(Poly3DCollection(verts, alpha=0.2))

        ax.set_xlim(-1, 1)
        ax.set_ylim(-1, 1)
        ax.set_zlim(-1, 1)

        ax.set_xlabel("Easting")
        ax.set_ylabel("Northing")
        ax.set_zlabel("Elevation")

        ax.legend()

        return fig, ax

    def plot_strip(self, line, width=100, cloud=None, image=None, s=2, aspect='auto'):
        """
        Plot a projected strip for comparision between an image and a point cloud.

        Args:
            line: the line ID to plot.
            width: how many pixels to plot either side of the line. Default is 10.
            cloud: the cloud to plot. Can be None.
            image: the (linescanner) image to plot. Can be None.
            s: size of rendered points in pixels. Default is 2.
            aspect: the aspect ratio of the renders / image plot. Default is 'equal'. Change to
                    'auto' to stretch data to figure size.
        """

        # build plot
        if cloud is None and image is None:
            assert False, "At least one dataset (cloud or image) must be passed for plotting."
        elif cloud is None or image is None:
            fig, ax = plt.subplots(1, 1, figsize=(15, 5))
            iax = cax = ax
        else:
            fig, ax = plt.subplots(2, 1, figsize=(15, 5))
            cax = ax[0]
            iax = ax[1]

        # plot cloud
        if cloud is not None:
            # project onto frame
            pp = self.project_to_frame(cloud, line)
            pp[:, 1] += width / 2  # offset origin to left side of image

            # calculate point visibility (including points within width)
            vis = (pp[:, 0] > 0) & (pp[:, 0] < self.dims[0]) & (pp[:, 1] > -(width / 2)) & (pp[:, 1] < width / 2)
            # pp[:, 0] += width / 2 # offset so we see either side of scan line

            # calculate point visibility (including points within width)
            vis = (pp[:, 0] > 0) & (pp[:, 0] < self.dims[0]) & (pp[:, 1] > 0) & (pp[:, 1] < width)

            # rasterise
            grd, z = rasterize(pp, vis, cloud.rgb, dims=(self.dims[0], width), s=s)

            # plot
            cax.imshow(np.transpose(grd, (1, 0, 2)) / 255.)
            cax.set_yticks([width / 2])
            cax.set_yticklabels(["line %d" % line])
            cax.set_aspect(aspect)
            cax.set_xticks([])
            cax.set_title("Point cloud RGB")

        if image is not None:
            image.quick_plot((0, 1, 2), ax=ax[1])
            iax.set_ylim(line - width, line + width)
            iax.set_yticks([line])
            iax.set_yticklabels(["line %d" % line])
            iax.set_aspect(aspect)
            iax.set_title("Scanner frame")

        fig.tight_layout()
        fig.show()


def project_pushbroom(image, cloud, cam, chunk=500, step=100, near_clip=10., vb=True):
    """
    Map an image acquired using a moving pushbroom scanner onto a point cloud using known
    camera position and orientations for each line of the image.

    Args:
        image: a HyImage instance containing the data to project. This is only used to determine image dimensions.
        cloud: the destination point cloud to project data onto.
        cam: a pushbroom camera instance.
        chunk: The size of chunks used to optimise the projection step. Points are culled based on the first and
               last line of each chunk prior to processing to reduce the number of projections that need to be
               performed. To reduce errors at chunk margins these chunks are padded by 50%. Default is 500.
        step: the step to use in the masking step for each chunk. Default is 100. Reducing this will ensure no points
              are missed, but at large performance cost.
        near_clip: the z-depth of the near clipping plane. Default is 10.0 m.
        vb: True if a progress bar should be displayed. Default is True.


    Returns:
        A hylite.project.PMap instance containing the mapping matrix between points and pixels.
    """

    # check dims match
    assert image.xdim() == cam.dims[0] \
           and image.ydim() == cam.dims[1], \
        "Error - image and camera dimensions do not match. Try rotating the image using image.rot90()."

    # store point IDs (so these don't get ruined by clipping)
    pointIDs = np.arange(0, cloud.point_count(), dtype=np.uint32)

    # build an adjacency matrix in flat form
    points = []
    pixels = []
    depths = []

    # loop through each chunk
    for c in range(0, image.ydim(), chunk):
        #######################################################
        ## Clip point cloud to chunk using coarse projections
        #######################################################
        # calculate start of frame projection and facing
        P0 = cam.project_to_frame(cloud, c)
        F0 = P0[:, 1] > 0  # true for points in front of scan line
        mask = np.full(cloud.point_count(), False)  # mask of points visible in chunk
        loop = range(c + step, c + chunk + step, step)
        if vb:
            loop = tqdm(loop, leave=False, desc="Masking chunk %d/%d" % (c / chunk + 1, image.ydim() / chunk))
        for i in loop:
            if i >= image.ydim():  # reached end of image
                continue  # use continue here rather than break to use up the rest of the progress bar so it is removed

            P1 = cam.project_to_frame(cloud, i)
            F1 = P1[:, 1] > 0  # true for points in front of scan line
            vis = (F0 != F1)  # points visible in this frame
            vis = vis & np.logical_or((P0[:, 0] > 0) & (P0[:, 0] < cam.dims[0]),
                                      (P1[:, 0] > 0) & (P1[:, 0] < cam.dims[0]))
            mask[vis] = True  # these points should be included in final projection
            F0 = F1
            P0 = P1

        # subset points to those visible in this chunk (only)
        xyz = cloud.xyz[mask]  # get points
        pIDs = pointIDs[mask]  # store IDs in the original point cloud

        #######################################################
        ## Do per-line projections for subset point cloud
        #######################################################
        # calculate start of frame projection and facing (with subsetted cloud)
        P0 = cam.project_to_frame(xyz, c)
        F0 = P0[:, 1] > 0  # true for points in front of scan line

        loop = range(c + 1, c + chunk + 1)
        if vb:
            loop = tqdm(loop, leave=False,
                        desc="Projecting chunk %d/%d (%d points)" % (c / chunk + 1, image.ydim() / chunk, xyz.shape[0]))
        for i in loop:
            if i >= image.ydim():  # reached end of image
                continue  # use continue here rather than break to use up the rest of the progress bar so it is removed

            # calculate end of frame projection and facing
            P1 = cam.project_to_frame(xyz, i)
            F1 = P1[:, 1] > 0  # true for points in front of scan line

            # get points crossed by pushbroom
            vis = (F0 != F1)  # N.B. this culls most of the points in the cloud!

            # remove points out of field of view (but co-planar with line)
            vis = vis & np.logical_or((P0[:, 0] > 0) & (P0[:, 0] < cam.dims[0]),
                                      (P1[:, 0] > 0) & (P1[:, 0] < cam.dims[0])) \
                  & np.logical_or(P0[:, 2] > near_clip, P1[:, 2] > near_clip)

            # cull invisible points
            _PP = np.vstack([np.clip(P0[vis, 0], 0, image.xdim() - 1),
                             np.clip(P1[vis, 0], 0, image.xdim() - 1)]).T  # cast to indices
            _PP = np.sort(_PP, axis=1).astype(np.uint)  # sort so _PP[i,0]:_pp[i,1] gives the range of indices.
            _PP[:, 1] += 1  # add one to final index for slicing
            _Z = np.max([P0[vis, 2], P1[vis, 2]], axis=0)
            for n, pid in enumerate(pIDs[vis]):
                # add to link matrix
                px = range(int(i * image.xdim() + _PP[n, 0]), int(i * image.xdim() + _PP[n, 1]))
                pixels += px
                points += [pid] * len(px)
                depths += [_Z[n]] * len(px)

            # update left clipping plane
            F0 = F1
            P0 = P1

    # build a projection map and return it
    pmap = PMap(image.xdim(), image.ydim(), cloud.point_count(), cloud=cloud, image=image)
    pmap.set_flat(points, pixels, np.array(depths))

    return pmap

def get_corr_coef(pmap, bands=(0, 1, 2)):
    """
    Calculate the pearsons correlation coefficient between projected RGB colours (pmap.image[...,bands])
    and pmap.cloud.rgb.  Used to optimise boresight values against point
    cloud colours (slow, but worthwhile).

    Args:
        pmap: the pmap instance containing the projection.
        bands: indices for the red, green and blue bands of pmap.image. Default is (0,1,2).

    Returns:
        the corellation of the red, green and blue bands.
    """
    assert pmap.cloud.rgb is not None, "Error - cloud must contain independent RGB information."
    rgb_ref = pmap.cloud.rgb  # 'true' RGB from SfM
    rgb_adj = push_to_cloud(pmap, bands, method='count').data  # calculate projected RGB

    # filter to known points
    mask = (rgb_adj > 0).any(axis=-1) & (rgb_ref > 0).any(axis=-1)

    # return the corellation of the red, green and blue bands.
    return [sp.stats.pearsonr(rgb_ref[mask, i], rgb_adj[mask, i])[0] for i in range(3)]


def optimize_boresight(track, cloud, image, bands=(0, 1, 2), n=100, iv=np.array([0, 0, 0]), scale=3.,
                       coarse_eps=0.5, fine_eps=0.05, ztol=0, ftol=0,
                       vb=True, gf=True):
    """
    Applies a least-squares solver to find boresight values that result in the best correlation
    between projected RGB and RGB stored on a photogrammetric point cloud. This can be very slow,
    so use the subsample argument to significantly subsample the point cloud. It is also a good idea
    to calculate these values using a small but high-quality subset of a swath; the optimised values
    can subsequently be applied to the whole dataset

    Args:
        cloud: the cloud containing the geometry to project onto and associated RGB values.
        image: the image containing RGB data.
        n: the subsampling factor for the pointcloud (uses cloud.xyz[::n,:]). Default is 100.
        iv: the initial value for the boresight estimation. Default is [0,0,0].
        bands: indices for the red, green and blue bands of image. Default is (0,1,2).
        scale: The size of the search space. Default is ±3 degrees from the initial value.
        coarse_eps: the perturbation step size for the first solution.
        fine_eps: the perturbation step size for subsequent refinement.
        ztol: the tolerance of the z-filtering step. Set to 0 to disable (default).
        ftol: the maximum number of points per (valid) pixel. Set to 0 to disable (default).
        vb: True if output of the least-squares solver should be printed at each state.
        gf: True if a plot should be generated showing the search trace. Default is True

    Returns:
        A tuple containing:

         - track = an updated track with the boresight applied.
         - boresight = a (3,) array with optimised boresight adjustments (roll, pitch, yaw).
         - trace = the search path of the optimisation, with shape (niter,4). The last column contains
                    the cost function at each iteration point.

    """

    # check cloud and build subsampled copy
    assert cloud.rgb is not None, "Error - cloud must contain independent RGB information."
    cloud = hylite.HyCloud(cloud.xyz[::n, :], rgb=cloud.rgb[::n, :])

    # setup cost function
    trace = []

    def _opt(X, track, cloud, image):
        # update track
        track = track.apply_boresight(*X)

        # build projection map
        pmap = project_pushbroom(image, cloud, track, chunk=200, step=50, vb=False)

        # filter
        if ztol > 0:
            pmap.filter_occlusions( ztol )
        if ftol > 0:
            pmap.filter_footprint( ftol )

        # return cost
        c = -np.sum(get_corr_coef(pmap, bands=bands))
        trace.append((X[0], X[1], X[2], -c))  # add to trace
        if vb:
            print("\r%d: boresight = [%.3f,%.3f,%.3f]; correlation=%.4f." % (len(trace), X[0], X[1], X[2], -c / 3),
                  end='  ')

        return c

    # calculate bounds
    bounds = [(v - scale, v + scale) for v in iv]

    # run coarse optimisation
    res = minimize(_opt, np.array(iv), args=(track, cloud, image),
                   bounds=bounds, method='TNC', options=dict(eps=coarse_eps))
    if vb:
        print("Complete [%s]." % res.message)

    tr = np.array(trace)
    s = np.argmax(
        tr[:, -1])  # get best value from iterations (with TNC this isn't always the optimal solution for some reason?)

    # run fine optimisation
    if fine_eps > 0:
        if vb:
            print("Running fine adjustment...")
        bounds = [(v - scale, v + scale) for v in iv]
        res = minimize(_opt, tr[s, :3], args=(track, cloud, image),
                       bounds=bounds, method='TNC', options=dict(eps=fine_eps))
        if vb:
            print("Complete [%s]." % res.message)
        tr = np.array(trace)

    tr[:, -1] /= 3  # convert sum of corellation coeffs to average (is more interpretable)
    s = np.argmax(
        tr[:, -1])  # get best value from iterations (with TNC this isn't always the optimal solution for some reason?)

    # plot?
    if gf:
        plt.plot(tr[:, 0], tr[:, 3], color='r', label='Roll')
        plt.plot(tr[:, 1], tr[:, 3], color='g', label='Pitch')
        plt.plot(tr[:, 2], tr[:, 3], color='b', label='Yaw')
        plt.scatter(tr[s, 0], tr[s, 3], color='k', zorder=10)
        plt.scatter(tr[s, 1], tr[s, 3], color='k', zorder=10)
        plt.scatter(tr[s, 2], tr[s, 3], color='k', zorder=10)
        plt.title("Iter %d: boresight = [%.3f,%.3f,%.3f]" % (tr.shape[0], tr[s, 0], tr[s, 1], tr[s, 2]))
        plt.legend()
        plt.xlabel('Value')
        plt.ylabel('Correlation')
        plt.show()

    # apply to track
    return track.apply_boresight(*tr[s, :3]), tr[s, :3], tr


"""
Combine 2D image data and 3D point cloud information in one data structure that facilitates transfer of information
between images and point clouds.
"""

from hylite import HyCollection
from hylite.project import PMap, Pushbroom, Camera, proj_persp, proj_pano, project_pushbroom, push_to_cloud, push_to_image
from tqdm import tqdm
from scipy.ndimage import grey_dilation
import numpy as np

class HyScene( HyCollection ):
    """
    A special type of HyCollection and contains a projection map for transferring information between a point cloud and
    a hyperspectral image.
    """
    def __init__(self, name, root, header=None ):
        """
        Args:
            name (str): unique name for this scene
            root (str): path to folder where scene will be saved to disk
            header (str): header information for this scene, or None.
        """

        super().__init__( name, root, header )
        self.ext = '.hys'

    def getAttributes(self):
        """
        Return a list of available attributes in this HyScene. We must override the HyCollection implementation to remove
        functions associated with HyScene.
        """
        return list(set(dir(self)) - set(dir(HyCollection)) - set(dir(HyScene)) - set(['header', 'root', 'name', 'ext']))

    def construct(self, image, cloud, camera, s=1, occ_tol = 10, maxf=0, bf=True, vb = True, **kwds):
        """
        Construct a mapping between the specified image and cloud based on the camera position / orientation / track.

        Args:
            image (HyImage): data to map onto the point cloud.
            cloud (HyCloud): data/geometry to map onto the image.
            camera (hylite.project.Camera or hylite.project.Pushbroom):  Object describing the projection geometry.
            s (int): a dilation to apply when mapping point data to the image (to fill gaps/holes). Default is 1 (do not apply a dilation).
                If s is an integer then an (s,s) dilation filter is applied. Alternatively, s can be a tuple such that s=(n,m)
                defining the 2-D dimensions of the dilation (useful for e.g. pushbroom data).
            occ_tol (float): the distance between a point and the z-buffer before it becomes occluded. Default is 10. Set to 0 to
                    disable occlusion.
            maxf (int): the maximum acceptible pixel footprint. Pixels containing > than this number of points will be excluded
                   from the dataset. Set as 0 to disable (default).
            bf (bool): True if backface culling (using cloud normal vectors) should be applied during projection. Default is True.
            **kwds: Keywords are passed to project_pushbroom for pushbroom type cameras.
        """

        # check dimensions match
        assert camera.dims[0] == image.xdim() and camera.dims[
            1] == image.ydim(), "Error - image and camera dimensions do not match."

        # store reference to associated datasets and properties
        self.image = image
        self.cloud = cloud
        self.camera = camera
        self.occ_tol = occ_tol
        self.maxf = maxf

        # build projection map
        self.pmap = PMap(camera.dims[0], camera.dims[1], cloud.point_count(), cloud=cloud, image=image)
        normals = None
        if bf:
            normals = cloud.normals
        if isinstance(self.camera, Camera): # conventional camera
            if 'persp' in camera.proj:
                pp, vis = proj_persp(cloud.xyz, C=camera.pos, a=camera.ori,
                                     fov=camera.fov, dims=camera.dims, normals=normals)
                self.pmap.set_ppc( pp, vis )
            elif 'pano' in camera.proj:
                pp, vis = proj_pano(cloud.xyz, C=camera.pos, a=camera.ori,
                                    fov=camera.fov, dims=camera.dims, step=camera.step, normals=normals)
                self.pmap.set_ppc(pp, vis)
            else:
                assert False, "Error, %s is an incompatible projection type" % camera.proj
        elif isinstance(self.camera, Pushbroom):
                self.pmap = project_pushbroom( self.image, self.cloud, self.camera, vb=vb, **kwds )
        else:
            assert False, "Error, %s is an unknown camera type" % type(self.camera)


        # filter occlusions
        if vb:
            prg = tqdm(total=6, leave=False)
            prg.set_description("Filtering occlusions")
        if self.occ_tol > 0:
            self.pmap.filter_occlusions( self.occ_tol )

        # filter footprint
        if maxf > 0:
            if vb:
                prg.set_description("Filtering by footprint")
                prg.update(1)
            self.pmap.filter_footprint( self.maxf )

        # do projections
        if vb:
            prg.set_description("Converting geometry")
            prg.update(2)
        if cloud.has_normals():
            xyzklm = push_to_image( self.pmap, 'xyzklm', method='average').data
            self.xyz = xyzklm[...,[0,1,2]]
            self.normals = xyzklm[..., [3, 4, 5]]
        else:
            self.xyz = push_to_image( self.pmap, 'xyz', method='average').data
            self.normals = None

        # build depth map
        if vb:
            prg.set_description("Building depth map")
            prg.update(3)
        self.depth = self.pmap.get_pixel_depths()

        # build viewvector map
        if vb:
            prg.set_description("Computing view vectors")
            prg.update(5)
        if isinstance( self.camera, Camera ): # fixed view position
            self.view = self.xyz - self.camera.pos
        else: # camera positions changes according to pushbroom track
            self.view = self.xyz - self.camera.cp
        self.view = self.view / np.linalg.norm(self.view, axis=-1)[...,None] # normalise

        # apply dilations
        if isinstance(s, tuple) or s > 1:
            if vb:
                prg.set_description("Applying dilations")
                prg.update(6)
            if isinstance(s,int):
                s = (s,s)
            elif isinstance(s, tuple):
                assert len(s) == 2, "Error - s must be a tuple of shape (n,m)."
            else:
                assert False, "Error, %s is an invalid type for s." % type(s)

            # loop through attributes and apply dilation
            for a,v in zip(['xyz','normals','depth', 'view'], [self.xyz, self.normals, self.depth, self.view]):
                nanmask = np.logical_not(np.isfinite(v))
                v[nanmask] = -np.inf  # replace nans with -infinity
                if (len(v.shape) == 2):
                    k = np.ones(s)
                elif (len(v.shape) == 3):
                    k = np.ones( s + (1,))
                dil = grey_dilation(v, footprint=k)
                v[nanmask] = dil[nanmask]  # where we have new finite values from dilation, add them in
                v[np.isinf(v)] = np.nan  # add nans back again



        if vb:
            prg.set_description("Complete")
            prg.close()

    def get_pixel_normal(self, px, py):
        """
        Get the average normal vector of all points in the specified pixel.
        """
        assert self.cloud.has_normals(), "Error - no normals are available (on point cloud)."
        return self.normals[px,py,:]

    def get_point_normal(self, index):
        """
        Get the normal vector of the specified point.
        """
        assert self.cloud.has_normals(), "Error - no normals are available (on point cloud)."
        return self.cloud.normals[index,:]

    def get_xyz(self):
        """
        Get per-pixel positions in world coords (as a numpy array)
        """
        return self.xyz

    def get_normals(self):
        """
        Get per-pixel normals (as a numpy array).
        """
        return self.normals

    def get_depth(self):
        """
        Get per-pixel depth array.
        """
        return self.depth

    def get_GSD(self):
        """
        Get per-pixel ground sampling distance (pixel size). Assumes square pixels.

        Returns:
            gsd (numpy array): sampling distance (GSD) in meters.
        """
        # calculate pixel pitch in degrees
        if isinstance(self.camera, Camera): # normal camera
            pitch = np.deg2rad(self.camera.dims[1] / self.camera.fov)
        else: # pushbroom camera
            pitch = np.deg2rad( 1 / self.camera.xfov )

        # calculate GSD and return
        return 2 * self.depth * np.tan(pitch / 2)

    def get_obliquity(self):
        """
        Get array obliquity angles (degrees) between camera look direction and surface normals. Note that this angle
        will be 0 if the look direction is perpendicular to the surface.
        """
        return np.dot(self.view, self.normals)

    def get_view_dir(self):
        """
        Get per-pixel viewing direction vector (normalised to length = 1).
        """
        return self.view

    def get_slope(self):
        """
        Get array of slope angles for each pixel (based on the surface normal vectors).
        """
        return np.rad2deg(np.arccos(np.abs(self.normals[..., 2])))

    def push_to_cloud(self, bands, method='best', image=None, cloud=None):
        """
        Push attributes from this scenes image to this scene's cloud. Is a wrapper around hylite.project.push_to_cloud(...).

        Args:
            bands (int, float, tuple, list): either:
                     (1) a index (int), wavelength (float) of a (single) image band to export.
                     (2) a tuple containing the (min,max) wavelength to extract. If range is a tuple, -1 can be used to specify the
                         first or last band index.
                     (3) a list of bands or boolean mask such that image.data[:,:,range] is exported.
            method (string): The method used to condense data from multiple pixels onto each point. Options are:

                             - 'closest': use the closest pixel to each point.
                             - 'distance': average with inverse distance weighting.
                             - 'count' : average weighted inverse to the number of points in each pixel.
                             - 'best' : use the pixel that is mapped to the fewest points (only). Default.
                             - 'average' : average with all pixels weighted equally.

            image (hylite.HyImage): an alternative image to use (defaults to self.image) Shapes must match self.pmap.
            cloud (hylite.HyCloud): an alternative cloud to use (defaults to self.cloud). Shapes must match self.pmap.

        Returns:
            A HyCloud instance containing the back-projected data.
        """
        if cloud is None:
            cloud = self.cloud
        if image is None:
            image = self.image

        return push_to_cloud( self.pmap, bands, method, image=image, cloud=cloud)

    def push_to_image(self, bands, method='closest', image=None, cloud=None):
        """
        Push attributes from this scenes cloud to this scenes image. Is a wrapper around hylite.project.push_to_image(...).

        Args:
            bands (int,float,str,tuple,list): List defining the bands to include in the output dataset. Elements should be one of:
                  - numeric = index (int), wavelength (float) of an image band
                  - bands = a list of image band indices (int) or wavelengths (float). Inherent properties of point clouds
                       can also be expected by passing any combination of the following:

                        - 'rgb' = red, green and blue per-point colour values
                        - 'klm' = point normals
                        - 'xyz' = point coordinates

                  - iterable of length > 2: list of bands (float or integer) to export.
            method (str): The method used to condense data from multiple points onto each pixel. Options are:

                         - 'closest': use the closest point to each pixel (default is this is fastest).
                         - 'average' : average with all pixels weighted equally. Slow.

            image (hylite.HyImage): an alternative image to use (defaults to self.image) Shapes must match self.pmap.
            cloud (hylite.HyCloud): an alternative cloud to use (defaults to self.cloud). Shapes must match self.pmap.

        Returns:
            A HyImage instance containing the projected data.
        """
        if cloud is None:
            cloud = self.cloud
        if image is None:
            image = self.image

        return push_to_image( self.pmap, bands, method, image=image, cloud=cloud )

    # def match_colour_to(self, reference, uniform=True, method='norm', inplace=True):
    #
    #     """
    #     Identifies matching pixels between two hyperspectral scenes and uses them to minimise
    #     colour differences using a linear model (aka by adjusting the brightness/contrast of this scene
    #     to match the brightness/contrast of the reference scene). WARNING: by default this modifies this scene's
    #     image IN PLACE.
    #
    #     Args:
    #     - reference = the scene to match colours to.
    #     - uniform = True if a single brightness contrast adjustment is applied to all bands (avoids introducing spectral
    #              artefacts). If False, different corrections area applied to each band - use with CARE! Default is True.
    #     - method = The colour matching method to use. Current options are:
    #                 - 'norm' = centre-means and scale to match standard deviation. Only compares points known to match.
    #                 - 'hist' = histogram equalisation. Applies to all pixels in scene - use with care!
    #                Default is 'norm'.
    #     - inplace = True if the correction should be applied to self.image in-place. If False, no correction is
    #               applied, and the correction weights (cfac and mfac) returned for future use. Default is True.
    #     Returns:
    #     - The corrected image as a HyImage object. If inplace=True (default) then this will be the same as self.image.
    #     """
    #
    #     image = self.image
    #     if not inplace:
    #         image = image.copy()
    #
    #     if 'norm' in method.lower():
    #         # get matching pixels
    #         px1, px2 = self.intersect_pixels(reference)
    #         assert px1.shape[0] > 0, "Error - no overlap between images."
    #         if px1.shape[0] < 1000:
    #             print("Warning: images only have %d overlapping pixels,"
    #                " which may result in poor colour matching." % px1.shape[0])
    #
    #         # extract data to create vector of matching values
    #         px1 = image.data[px1[:, 0], px1[:, 1], :]
    #         px2 = reference.image.data[px2[:, 0], px2[:, 1], :]
    #
    #         # apply correction
    #         image.data = norm_eq( image.data, px1, px2, per_band=not uniform, inplace=True)
    #
    #     elif 'hist' in method.lower():
    #         if uniform: # apply to whole dataset
    #             image.data = hist_eq(image.data, reference.image.data)
    #         else: # apply per band
    #             for b in range(self.image.band_count()):
    #                 image.data[:, :, b] = hist_eq(image.data[:, :, b], reference.image.data[:, :, b])
    #     else:
    #         assert False, "Error - %s is an unrecognised colour correction method." % method
    #
    #     return image
    #
    # ###################################
    # ##PLOTTING AND EXPORT FUNCTIONS
    # ###################################
    # def _gather_bands(self, bands):
    #     """
    #     Utility function used by push_to_image( ... ) and push_to_cloud( ... ).
    #     """
    #
    #     # extract wavelength and band name info
    #     wav = []
    #     nam = []
    #
    #     # loop through bands tuple/list and extract data indices/slices
    #     for e in bands:
    #         # extract from point cloud based on string
    #         if isinstance(e, str):
    #             for c in e.lower():
    #                 if c == 'r':
    #                     assert self.cloud.has_rgb(), "Error - RGB information not found."
    #                     nam.append('r')
    #                     wav.append(hylite.RGB[0])
    #                 elif c == 'g':
    #                     assert self.cloud.has_rgb(), "Error - RGB information not found."
    #                     nam.append('g')
    #                     wav.append(hylite.RGB[1])
    #                 elif c == 'b':
    #                     assert self.cloud.has_rgb(), "Error - RGB information not found."
    #                     nam.append('b')
    #                     wav.append(hylite.RGB[2])
    #                 elif c == 'x':
    #                     nam.append('x')
    #                     wav.append(-1)
    #                 elif c == 'y':
    #                     nam.append('y')
    #                     wav.append(-1)
    #                 elif c == 'z':
    #                     nam.append('z')
    #                     wav.append(-1)
    #                 elif c == 'k':
    #                     assert self.cloud.has_normals(), "Error - normals not found."
    #                     nam.append('k')
    #                     wav.append(-1)
    #                 elif c == 'l':
    #                     assert self.cloud.has_normals(), "Error - normals not found."
    #                     nam.append('l')
    #                     wav.append(-1)
    #                 elif c == 'm':
    #                     assert self.cloud.has_normals(), "Error - normals not found."
    #                     nam.append('m')
    #                     wav.append(-1)
    #         # extract slice (from image)
    #         elif isinstance(e, tuple):
    #             assert len(e) == 2, "Error - band slices must be tuples of length two."
    #             idx0 = self.image.get_band_index(e[0])
    #             idx1 = self.image.get_band_index(e[1])
    #             if self.image.has_band_names():
    #                 nam += [self.image.get_band_names()[b] for b in range(idx0, idx1)]
    #             else:
    #                 nam += [str(b) for b in range(idx0, idx1)]
    #             if self.image.has_wavelengths():
    #                 wav += [self.image.get_wavelengths()[b] for b in range(idx0, idx1)]
    #             else:
    #                 wav += [float(b) for b in range(idx0, idx1)]
    #         # extract band based on index or wavelength
    #         elif isinstance(e, float) or isinstance(e, int):
    #             b = self.image.get_band_index(e)
    #             if self.image.has_band_names():
    #                 nam.append(self.image.get_band_names()[b])
    #             else:
    #                 nam.append(str(b))
    #             if self.image.has_wavelengths():
    #                 wav.append(self.image.get_wavelengths()[b])
    #             else:
    #                 wav.append(float(b))
    #         else:
    #             assert False, "Unrecognised band descriptor %s" % b
    #
    #     return wav, nam
    #
    # def push_to_image(self, bands, fill_holes=False, blur=0):
    #     """
    #     Export data from associated cloud and image to a (new) HyImage object.
    #
    #     Args:
    #      - bands = a list of image band indices (int) or wavelengths (float). Inherent properties of point clouds
    #                can also be expected by passing any of the following:
    #                 - 'rgb' = red, green and blue per-point colour values
    #                 - 'klm' = point normals
    #                 - 'xyz' = point coordinates
    #      - fill_holes = post-processing option to fill single-pixel holes with maximum value from adjacent pixels. Default is False.
    #      - blur = size of gaussian kernel to apply to image in post-processing. Default is 0 (no blur).
    #     Returns:
    #      - a HyImage object containing the requested data.
    #     """
    #
    #     # special case: individual band; wrap in list
    #     if isinstance(bands, int) or isinstance(bands, float) or isinstance(bands, str):
    #         bands = [bands]
    #
    #     # special case: tuple of two bands; treat as slice
    #     if isinstance(bands, tuple) and len(bands) == 2:
    #         bands = [bands]
    #
    #     # gather bands and extract wavelength and name info
    #     wav, nam = self._gather_bands(bands)
    #
    #     # rasterise and make HyImage
    #     img = np.full((self.image.xdim(), self.image.ydim(), len(wav)), np.nan)
    #     for _x in range(self.image.xdim()):
    #         for _y in range(self.image.ydim()):
    #             if not self.valid[_x, _y]:
    #                 continue
    #             pID = self.get_point_index(_x, _y)
    #             n = 0
    #             for e in bands:
    #                 if isinstance(e, str):  # extract from point cloud based on string
    #                     for c in e.lower():
    #                         if c == 'r':
    #                             img[_x, _y, n] = self.cloud.rgb[pID, 0]
    #                         elif c == 'g':
    #                             img[_x, _y, n] = self.cloud.rgb[pID, 1]
    #                         elif c == 'b':
    #                             img[_x, _y, n] = self.cloud.rgb[pID, 2]
    #                         elif c == 'x':
    #                             img[_x, _y, n] = self.cloud.xyz[pID, 0]
    #                         elif c == 'y':
    #                             img[_x, _y, n] = self.cloud.xyz[pID, 1]
    #                         elif c == 'z':
    #                             img[_x, _y, n] = self.cloud.xyz[pID, 2]
    #                         elif c == 'k':
    #                             img[_x, _y, n] = self.normals[_x, _y, 0]
    #                         elif c == 'l':
    #                             img[_x, _y, n] = self.normals[_x, _y, 1]
    #                         elif c == 'm':
    #                             img[_x, _y, n] = self.normals[_x, _y, 2]
    #                         n += 1
    #                     continue
    #                 elif isinstance(e, tuple):  # extract slice (from image)
    #                     assert len(e) == 2, "Error - band slices must be tuples of length two."
    #                     idx0 = self.image.get_band_index(e[0])
    #                     idx1 = self.image.get_band_index(e[1])
    #                     slc = self.image.data[_x, _y, idx0:idx1]
    #                     img[_x, _y, n:n + len(slc)] = slc
    #                     n += len(slc)
    #                     continue
    #                 elif isinstance(e, float) or isinstance(e, int):  # extract band based on index or wavelength
    #                     b = self.image.get_band_index(e)
    #                     img[_x, _y, n] = self.image.data[_x, _y, b]
    #                     n += 1
    #                     continue
    #                 else:
    #                     assert False, "Unrecognised band descriptor %s" % b
    #
    #     # build HyImage
    #     img = HyImage(img, header=self.image.header.copy())
    #     img.set_band_names(nam)
    #     img.set_wavelengths(wav)
    #
    #     # postprocessing
    #     if fill_holes:
    #         img.fill_holes()
    #     if blur > 2:
    #         img.blur(int(blur))
    #
    #     return img
    #
    # def push_to_cloud(self, bands):
    #     """
    #     Export data from associated image and cloud to a (new) HyCloud object.
    #
    #     Args:
    #      - bands = a list of image band indices (int) or wavelengths (float). Inherent properties of point clouds
    #                can also be expected by passing any of the following:
    #                 - 'rgb' = red, green and blue per-point colour values
    #                 - 'klm' = point normals
    #                 - 'xyz' = point coordinates
    #     Returns:
    #      - a HyImage object containing the requested data.
    #     """
    #
    #     # special case: individual band; wrap in list
    #     if isinstance(bands, int) or isinstance(bands, float) or isinstance(bands, str):
    #         bands = [bands]
    #
    #     # special case: tuple of two bands; treat as slice
    #     if isinstance(bands, tuple) and len(bands) == 2:
    #         bands = [bands]
    #
    #     # gather bands and extract wavelength and name info
    #     wav, nam = self._gather_bands(bands)
    #
    #     # loop through points in cloud and add data
    #     data = np.full((self.cloud.point_count(), len(wav)), np.nan)
    #     valid = np.full(self.cloud.point_count(), False, dtype=np.bool)
    #     for i in range(self.cloud.point_count()):
    #         # is point visible?
    #         _x, _y = self.get_pixel(i)
    #         if _x is None:
    #             continue
    #
    #         valid[i] = True  # yes - this point has data
    #
    #         # gather data
    #         n = 0
    #         for e in bands:
    #             if isinstance(e, str):  # extract from point cloud based on string
    #                 for c in e.lower():
    #                     if c == 'r':
    #                         data[i, n] = self.cloud.rgb[i, 0]
    #                     elif c == 'g':
    #                         data[i, n] = self.cloud.rgb[i, 1]
    #                     elif c == 'b':
    #                         data[i, n] = self.cloud.rgb[i, 2]
    #                     elif c == 'x':
    #                         data[i, n] = self.cloud.xyz[i, 0]
    #                     elif c == 'y':
    #                         data[i, n] = self.cloud.xyz[i, 1]
    #                     elif c == 'z':
    #                         data[i, n] = self.cloud.xyz[i, 2]
    #                     elif c == 'k':
    #                         data[i, n] = self.cloud.normals[i, 0]
    #                     elif c == 'l':
    #                         data[i, n] = self.cloud.normals[i, 1]
    #                     elif c == 'm':
    #                         data[i, n] = self.cloud.normals[i, 2]
    #                     n += 1
    #                 continue
    #             elif isinstance(e, tuple):  # extract slice (from image)
    #                 assert len(e) == 2, "Error - band slices must be tuples of length two."
    #                 idx0 = self.image.get_band_index(e[0])
    #                 idx1 = self.image.get_band_index(e[1])
    #                 slc = self.image.data[_x, _y, idx0:idx1]
    #                 data[i, n:(n + len(slc))] = slc
    #                 n += len(slc)
    #                 continue
    #             elif isinstance(e, float) or isinstance(e, int):  # extract band based on index or wavelength
    #                 b = self.image.get_band_index(e)
    #                 data[i, n] = self.image.data[_x, _y, b]
    #                 n += 1
    #                 continue
    #             else:
    #                 assert False, "Unrecognised band descriptor %s" % b
    #
    #     # build HyCloud
    #     cloud = self.cloud.copy(data=False)
    #     cloud.data = data
    #     cloud.filter_points(0, np.logical_not(valid))  # remove points with no data
    #     cloud.set_band_names(nam)
    #     cloud.set_wavelengths(wav)
    #
    #     return cloud
    #
    # def quick_plot(self, band=0, ax=None, bfac=0.0, cfac=0.0,
    #                **kwds):
    #     """
    #     Plot a projected data using matplotlib.imshow(...).
    #
    #     Args:
    #      - band = the band name (string), index (integer) or wavelength (float) to plot. Default is 0. If a tuple is passed then
    #               each band in the tuple (string or index) will be mapped to rgb.
    #      - bands = List defining the bands to include in the output image. Elements should be one of:
    #           - 'rgb' = rgb
    #           - 'xyz' = point position
    #           - 'klm' = normal vectors
    #           - numeric = index (int), wavelength (float) of an image band
    #           - tuple of length 3: wavelengths or band indices to map to rgb.
    #      - ax = an axis object to plot to. If none, plt.imshow( ... ) is used.
    #      - bfac = a brightness adjustment to apply to RGB mappings (-1 to 1)
    #      - cfac = a contrast adjustment to apply to RGB mappings (-1 to 1)
    #     *Keywords*:
    #      - keywords are passed to matplotlib.imshow( ... ).
    #     """
    #
    #     # plot hyImage data
    #     if not isinstance(band, str):
    #         kwds["mask"] = np.logical_not( np.isfinite(self.depth) ) #np.logical_not(self.valid)  # mask out pixels with no valid point mappings
    #         return self.image.quick_plot(band, ax, bfac, cfac, **kwds)  # render
    #     else:
    #         img = self.push_to_image(band)
    #         if (len(band) == 3):
    #             # do some normalizations
    #             mn = kwds.get("vmin", np.nanmin(img.data[img.data != 0]))
    #             mx = kwds.get("vmax", np.nanmax(img.data[img.data != 0]))
    #             img.data = (img.data - mn) / (mx - mn)
    #             if 'x' in band or 'y' in band or 'z' in band:  # coordinates need per-band mapping
    #                 for i in range(3):
    #                     img.data[..., i] = (img.data[..., i] - np.nanmin(img.data[..., i])) / (
    #                             np.nanmax(img.data[..., i]) - np.nanmin(img.data[..., i]))
    #             # plot it image
    #             return img.quick_plot((0, 1, 2), ax=ax, bfac=bfac, cfac=cfac, **kwds)
    #         else:
    #             return img.quick_plot(0, ax=ax, bfac=bfac, cfac=cfac, **kwds)
    #
    # @classmethod
    # def build_hypercloud(cls, scenes, bands, blending_mode='average', trim=True, vb=True, inplace=True, export_footprint = False):
    #     """
    #     Combine multiple HyScene objects into a hypercloud. Warning - this modifies the cloud associated with the HyScenes in-place.
    #     Returns processed cloud and (optional) footprint map indicating the number of involved scenes per pixel.
    #
    #     Args:
    #      - scenes = a list of scenes to combine. These scenes must all reference the same point cloud!
    #      - bands = either:
    #              (1) a tuple containing the (min,max) wavelength to map. If range is a tuple, -1 can be used to specify the
    #                  last band index.
    #              (2) a list of bands or boolean mask such that image.data[:,:,range] is exported to the hypercloud.
    #      - blending_mode = the mode used to blend points that can be assigned values from multiple values. Options are:
    #            - 'average' (default) = calculate the average of all pixels that map to the point
    #            - 'weighted' = calculate the average of all pixels, weighted by inverse footprint size.
    #            - 'gsd' = chose the pixel with the smallest gsd (i.e. the pixel with the closest camera)
    #            - 'obl' = chose the pixel that is least oblique (i.e. most perpendicular to the surface)
    #      - trim = True if the point cloud should be trimmed after doing project. Default is True.
    #      - vb = True if output should be written to the console. Default is True.
    #      - inplace = True if the reference cloud should be modified in place (to save RAM). Default is True.
    #      - export_footprint = True if footprint map indicating the number of involved scenes per pixel.
    #     """
    #
    #     if vb: print("Preparing data....", end='')
    #
    #     # get cloud
    #     cloud = scenes[0].cloud
    #     if not inplace:
    #         cloud = cloud.copy()
    #
    #     # remove everything except desired bands from scenes and calculate range of values
    #     images = []
    #     for i, s in enumerate(scenes):
    #         images.append(s.image.export_bands(bands))  # get bands of interest
    #         cloud.header.set_camera( s.camera, i ) # add camera to cloud header
    #
    #     # do scenes all have wavelength data?
    #     has_wav = np.array([i.has_wavelengths() for i in images]).all()
    #     wavelengths = None
    #     band_names = None
    #     if has_wav:  # match wavelengths between scenes
    #         wavelengths = []  # list of wavelengths
    #         indices = []  # list of corresponding band indices for each scene
    #         for w in images[0].get_wavelengths():
    #             idx = []
    #             for i, img in enumerate(images):
    #                 if not w in img.get_wavelengths():  # bail!
    #                     assert False, "Error - Scene %d does have data for data for band %d." % (i, w)
    #                 else:
    #                     idx.append(img.get_band_index(w))
    #             indices.append(idx)
    #             wavelengths.append(w)  # this wavelength is in all scenes
    #
    #         wavelengths = np.array(wavelengths)
    #         band_names = ["%.2f" % w for w in wavelengths]
    #         indices = np.array(indices)
    #
    #         assert wavelengths.shape[0] > 0, "Error - images do not have any matching bands."
    #     else:  # no - check all images have the same number of bands...
    #         assert (np.array([i.band_count() for i in images]) == images[0].band_count()).all(), \
    #             "Error - scenes with now wavelength information must have (exactly) the same band count."
    #         # no wavelengths, create band list and set band names
    #         band_list = np.array([i for i in range(scenes[0].image.band_count())])
    #         # noinspection PyUnusedLocal
    #         indices = np.array([band_list for s in scenes]).T  # bands to export
    #         # noinspection PyUnusedLocal
    #         band_names = ["%s" for b in band_list]
    #
    #     # handle bbl
    #     if np.array([img.header.has_bbl() for img in images]).all():
    #         bbl = np.full(wavelengths.shape[0], True)
    #         for i, idx in enumerate(indices):  # list of indices for each band
    #             bb = [images[n].header.get_bbl()[idx[n]] for n in range(len(images))]
    #             bbl[i] = np.array(bb).all()  # if any data is bad, this becomes a bad band
    #         cloud.header.set_bbl(bbl)  # store
    #
    #     if vb: print(" Done.")
    #
    #     # create data array
    #     data = np.zeros((cloud.point_count(), len(band_names)),
    #                     dtype=np.float32)  # point values will be accumulated here
    #     point_count = np.zeros(cloud.point_count(), dtype=np.float32)  # number of pixels used to calculate each point
    #
    #     # loop through points
    #     if vb:
    #         loop = tqdm(range(data.shape[0]), desc='Projecting data', leave=False)
    #     else:
    #         loop = range(data.shape[0])
    #     if 'average' in blending_mode.lower():
    #         for n in loop:
    #             for i, s in enumerate(scenes):
    #                 px, py = s.get_pixel(n)
    #                 if px is not None:  # point exists in scene
    #                     if np.isfinite(images[i].data[px, py, indices[:, i]]).all():  # is pixel finite?
    #                         point_count[n] += 1  # increment pixel count
    #                         data[n, :] += images[i].data[px, py, indices[:, i]]  # accumulate point value
    #         # divide by point count to calculate average
    #         data = data / point_count[:, None]
    #     elif 'gsd' in blending_mode.lower():
    #         for n in loop:
    #             best = np.inf
    #             for i, s in enumerate(scenes):
    #                 px, py = s.get_pixel(n)
    #                 if px is not None:  # point exists in scene
    #                     if np.isfinite(images[i].data[px, py, indices[:, i]]).all():  # is pixel finite?
    #                         point_count[n] += 1  # increment pixel count
    #                         if s.depth[px, py] < best:  # is this the best GSD so far?
    #                             best = s.depth[px, py]  # store best GSD
    #                             data[n, :] = images[i].data[px, py, indices[:, i]]  # set point value
    #     elif 'obl' in blending_mode.lower():
    #         for n in loop:
    #             best = np.inf
    #             for i, s in enumerate(scenes):
    #                 px, py = s.get_pixel(n)
    #                 if px is not None:  # point exists in scene
    #                     if np.isfinite(images[i].data[px, py, indices[:, i]]).all():  # is pixel finite?
    #                         point_count[n] += 1  # increment pixel count
    #                         if s.obliquity[px, py] < best:  # is this the best GSD so far?
    #                             best = s.obliquity[px, py]  # store best GSD
    #                             data[n, :] = images[i].data[px, py, indices[:, i]]  # set point value
    #     elif 'weight' in blending_mode.lower():
    #         for n in loop:
    #             for i, s in enumerate(scenes):
    #                 px, py = s.get_pixel(n)
    #                 if px is not None:  # point exists in scene
    #                     if np.isfinite(images[i].data[px, py, indices[:, i]]).all():  # is pixel finite?
    #                         w = 1 / s.depth[px, py] # calculate weight and increment to count
    #                         point_count[n] += w  #
    #                         data[n, :] += w * images[i].data[px, py, indices[:, i]]  # accumulate point value
    #         # divide by sum of weights to get weighted average
    #         data = data / point_count[:, None]
    #     else:
    #         assert False, "Error - unrecognised blending mode %s" % blending_mode
    #
    #     # add bands to point cloud
    #     cloud.set_bands(data, band_names=band_names, wavelengths=wavelengths)
    #
    #     # export footprint?
    #     if export_footprint:
    #         footprint = hylite.HyCloud(cloud.xyz, bands=point_count[:, None])
    #         if trim:
    #             footprint.filter_points(0, point_count == 0)  # remove zero points
    #
    #     # trim cloud?
    #     if trim:
    #         cloud.filter_points(0, point_count == 0)  # remove zero points
    #
    #     # return
    #     if export_footprint:
    #         return cloud, footprint
    #     else:
    #         return cloud






"""
Reference spectra such as calibration targets or spectral libraries.
"""

import numpy as np
import os
import glob
from pathlib import Path

class Target(object):
    """
    Class defining the reflectance properties of a calibration target.
    """
    def __init__(self, wavelength, reflectance, name = "Custom"):
        """
        Load a custom target from a .txt file with a band,reflectance pair on each line. Lines not of this format will be ignored.

        Args:
            wavelength (ndarray): the calibration wavelengths (in nanometers).
            reflectance (ndarray): the target reflectances.
            name (str): the name of this target.
        """

        self.band = np.array(wavelength)
        self.refl = np.array(reflectance)
        self.name = name

        assert self.band.shape[0] > 0, "Error: Calibration data must cover multiple wavelengths"
        assert self.band.shape == self.refl.shape, "Error: wavelength and reflectance data must have the same length."

    def get_name(self):
        """
        Return the name of this reference target.
        """
        return self.name

    def get_wavelengths(self):
        """
        Get the wavelengths corresponding to the reflectance measurements for this target.
        """
        return self.band

    def get_reflectance(self):
        """
        Get the reflectance measurements for this target.
        """
        return self.refl

    def writeTXT(self, path):
        """
        Save this calibration target to txt file
        """
        if os.path.isfile( path):
            path = os.path.dirname(path) #make sure path is a directory
        if not os.path.exists(path): #and it exists
            os.makedirs(path)

        #create file
        path = os.path.join( path, "%s.txt" % self.get_name())
        with open(path,'w') as f:
            w = self.get_wavelengths()
            r = self.get_reflectance()
            for i in range(len(w)):
                f.write("%f,%f\n" % (w[i],r[i]))

def loadTarget( path ):
    """
    Load a target from a .txt file with a band,reflectance pair on each line. Lines not of this format will be ignored.

    Args:
        path (str): path to the file to load.
    """

    band = []
    refl = []
    name = os.path.splitext(os.path.basename(path))[0]  # get name as filename
    with open(path, 'r') as f:
        lines = f.readlines()
        for l in lines:
            if ',' in l:
                split = l.split(",")  # comma delimited
            elif "\t" in l:
                split = l.split('\t')  # tab delimited
            else:
                split = l.split(' ')  # space delimited
            if len(split) == 2:
                band.append(float(split[0]))
                refl.append(float(split[1]))

    return Target( band, refl, name )

def loadDirectory( path ):
    """
    Load all spectra (stored as .txt files) in a directory.

    Args:
        path (str): the directory to search.

    Returns:
        A dictionary of reference (with target names as keys)
    """

    files = glob.glob( os.path.join(path, "*.txt" ) )
    targets = {}
    for f in files:
        try:
            t = loadTarget(f)
            targets[t.get_name()] = t
        except:
            print("Warning: could not parse target %s" % os.path.dirname(f))

    return targets


# load reference data from associated directories
spectralon = loadDirectory( str( Path(__file__).parent / "spectralon" ) )
PVC = loadDirectory( str( Path(__file__).parent / "pvc" ) )
custom = loadDirectory( str( Path(__file__).parent / "pvc" ) )

# use these to create local variables (for easy access/import)
locals().update(spectralon)
locals().update(PVC)
locals().update(custom)


"""
A package for applying radiometric and topographic corrections to hyperspectral datasets to convert measured at-sensor
radiance to reflectance. Data detrending methods (e.g. hull corrections) are also included here.
"""

from .detrend import get_hull_corrected
from .panel import Panel
from .equalize import norm_eq, hist_eq




import unittest
import os
from pathlib import Path
from hylite import io
import numpy as np

class MyTestCase(unittest.TestCase):
    def test_PCA_MNF(self):
        image = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"), "image.hdr"))
        cloud = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"), "hypercloud.hdr"))

        # run mnf on clouds and images
        from hylite.filter import MNF, PCA, from_loadings
        img_mask = np.full( (image.xdim(), image.ydim()), False )
        cld_mask = np.full( (cloud.point_count()), False )
        img_mask[:, :int(img_mask.shape[1]/2) ] = True
        cld_mask[:int(cloud.point_count() / 2)] = True
        for data, mask  in zip( [image, cloud], [img_mask, cld_mask]):

            # test PCA
            pca, w, m = PCA( data, bands=10, mask=mask)
            pca2 = from_loadings( data, w, m )
            self.assertEqual(pca.band_count(), 10)
            self.assertLess( np.nanmax( np.abs( pca.data - pca2.data ) ), 1e-4 )

            # run MNF
            mnf, w, m = MNF( data, bands=10, mask=mask)
            mnf2 = from_loadings(data, w, m)
            self.assertEqual( mnf.band_count(), 10 )
            self.assertLess(np.nanmax(np.abs(mnf.data - mnf2.data)), 1e-4)

            # run MNF denoising
            denoise, w, m = MNF( data, bands=10, denoise=True )
            self.assertEqual(denoise.band_count(), data.band_count())

        # test hull correction
        from hylite.correct import get_hull_corrected
        for d in [image, cloud]:
            hc = get_hull_corrected(d) # run hull correction
            self.assertLessEqual( np.nanmax(hc.data),  1 ) # assert results are not > 1
            self.assertLess( np.nanmin(hc.data), 1 ) # assert some values are less than one
            self.assertGreaterEqual( np.nanmin(hc.data), 0 ) # no negatives!
if __name__ == '__main__':
    unittest.main()



"""
Package containing spectral reference information for common minerals and calibration targets. Reference spectra are
located in the reference package while information on specific *absorbtion* features is stored in the *features* package.
"""

from .generate import randomSpectra, genImage